import uproot
import os
import shutil
import pandas as pd
import re
import matplotlib.pyplot as plt
from pathlib import Path

import torch
import torch.nn as nn

from opengate.base import GateObject
from opengate.exception import fatal
from opengate.utility import delete_folder_contents


def extract_number(filename):
    """
    Extracts the number from the filename.
    Helpful for sorting the csv files when fetched.
    """

    match = re.search(r"\d+", filename)
    if match:
        return int(match.group())
    return 0


def get_torch_device(gpu_mode):
    current_gpu_mode = None
    current_gpu_device = None
    if gpu_mode == "cpu":
        return torch.device("cpu")
    else:
        if torch.cuda.is_available():
            return torch.device("cuda")
        elif torch.backends.mps.is_available():
            return torch.device("mps")
        elif gpu_mode == "auto":
            # could not get a GPU device, so fall back to cpu
            return torch.device("cpu")
        else:
            fatal("GPU requested for torch, but no GPU on this device")


class WGANGenerator(nn.Module):
    """
    Generator class architecture for 3x3x3 crystal
    """

    def __init__(self, input_dim, output_dim, hidden_dim, labels_len):
        super(WGANGenerator, self).__init__()
        self.model = nn.Sequential(
            nn.Linear(input_dim + labels_len, hidden_dim),
            nn.ReLU(True),
            nn.Linear(hidden_dim, 2 * hidden_dim),
            nn.ReLU(True),
            nn.Linear(2 * hidden_dim, 4 * hidden_dim),
            nn.ReLU(True),
            nn.Linear(4 * hidden_dim, output_dim),
        )

    def forward(self, x):
        return self.model(x)


def _setter_hook_input_phsp_actor(self, input_phsp_actor):
    if input_phsp_actor is not None:
        if (
            self.simulation is not None
            and self.simulation != input_phsp_actor.simulation
        ):
            fatal(
                f"The input_phsp_actor {input_phsp_actor.name} refers to a different simulation than this {self.type_name}. "
            )
        self.simulation = input_phsp_actor.simulation
        self.root_file_path = input_phsp_actor.get_output_path()
    return input_phsp_actor


def process_root_output_into_events(df_combined):
    """
    This method takes the filtered root output (df_combined) and divides them
    into a list elements based on gamma event.

    All the info about one gamma event (gamma, electron, and optical photon info)
    is saved in one list element.
    """

    position_x = df_combined["Position_X"].to_numpy()
    position_y = df_combined["Position_Y"].to_numpy()
    position_z = df_combined["Position_Z"].to_numpy()
    particle_types = df_combined["ParticleType"].to_numpy()

    events = {}  # will store all the events
    current_event = []
    event_id = 0

    # flag to check if gamma is followed by electrons or photons
    gamma_has_electrons_or_photons = False

    # counts the occurrence of optical photons in current event
    optical_photon_count = 0

    # process each particle and segregate into events
    for index, (ptype, x, y, z) in enumerate(
        zip(particle_types, position_x, position_y, position_z)
    ):
        if ptype == "gamma":
            # store the previous event if it started with a gamma
            # and is followed by electrons or photons
            if current_event and gamma_has_electrons_or_photons:
                current_event.append(
                    {
                        "type": "opticalphoton",
                        "optical_photon_count": optical_photon_count,
                    }
                )
                events[event_id] = current_event
                event_id += 1
                optical_photon_count = 0
            # if it is a new event, add the gamma particle to it
            current_event = [{"index": index, "type": ptype, "x": x, "y": y, "z": z}]
            gamma_has_electrons_or_photons = False
        elif ptype == "opticalphoton":
            # if the particle is optical photon, just increment the count
            optical_photon_count += 1
            gamma_has_electrons_or_photons = True
        elif ptype == "e-":
            # Only add e- if there is an ongoing event (started by gamma)
            if current_event:
                current_event.append(
                    {"index": index, "type": ptype, "x": x, "y": y, "z": z}
                )
                gamma_has_electrons_or_photons = True

    # Store the last event if it is not empty
    if len(current_event) > 1:
        current_event.append(
            {"type": "opticalphoton", "optical_photon_count": optical_photon_count}
        )
        events[event_id] = current_event

    return events


class OptiGAN(GateObject):
    """
    Class Responsibilities:
    - Retrieve the necessary data from the root file.
    - Separate and organize data based on gamma events.
    - Load and initialize the OptiGAN model.
    - Extract and process inputs for use with OptiGAN.
    """

    user_info_defaults = {
        "output_folder": (
            "optigan",
            {
                "doc": "Folder where the OptiGAN saves output (some of it in subdirectories)."
                "It is relative to the simulation.output_dir if a simulation is known to OptiGAN, "
                "which is the case if the option 'input_phsp_actor' is set "
                "or if 'my_optigan.simulation=...' is set. "
                "Otherwise, output_folder is understood to be relative to the current working directory. "
                "output_folder can also be an absolute path."
            },
        ),
        "input_phsp_actor": (
            None,
            {
                "doc": "The phase space actor that generates the training data input for this OptiGAN. "
                "Its output path and the reference to the simulation are automatically picked up. ",
                "setter_hook": _setter_hook_input_phsp_actor,
            },
        ),
        "root_file_path": (
            None,
            {
                "doc": "Path to the root file containing training phase space data. "
                "This input is alternative to input_phsp_actor in case the OptiGAN is run stand-alone, "
                "i.e. without a GATE simulation.",
            },
        ),
        "path_to_optigan_model": (
            Path(os.path.dirname(__file__)) / "optigan_models" / "model_3341.pt",
            {
                "doc": "Path to the .pt model file used to create the GAN. "
                "The default model that currently ships with GATE 10 (model_3341.pt) "
                "is for 3*3*3 crystal dimension "
                "and has been trained by the team at UC Davis. ",
            },
        ),
        "torch_device": (
            "auto",
            {
                "doc": "The device to be used by torch. "
                "With 'auto', OptiGAN will try to get a GPU device "
                "and fall back to CPU if no GPU is available. ",
                "allowed_values": ("gpu", "cpu", "auto"),
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        # pick up the input_phsp_actor
        input_phsp_actor = kwargs.pop("input_phsp_actor", None)
        root_file_path = kwargs.pop("root_file_path", None)
        kwargs["name"] = "optigan"
        super().__init__(*args, **kwargs)
        # and set it once the base class has processed all user inout parameters
        if root_file_path is None and input_phsp_actor is None:
            fatal("An OptiGAN requires either a root_file_path or an input_phsp_actor.")

        # the output path from input_phsp_actor overrides the root_file_path
        self.root_file_path = root_file_path
        self.input_phsp_actor = input_phsp_actor

        # paths: they are set in initialize()
        self.optigan_input_folder = None
        self.optigan_output_folder = None
        self.optigan_csv_output_folder = None
        self.optigan_plots_folder = None

        self.events = {}
        self.extracted_events_details = []

        # Holds generator model
        self.generator = None

        # Arguments for the GAN.
        self.gan_arguments = {
            "noise_dimension": 10,  # input_dim to GAN
            "output_dimension": 6,
            "hidden_dimension": 128,
            "labels_length": 3,
        }

        self.device = None

    @property
    def _absolute_output_path(self):
        if self.simulation is not None:
            return self.simulation.get_output_path(
                self.output_folder, is_file_or_directory="directory"
            )
        else:
            return Path(".") / self.output_folder

    def get_absolute_path_to_folder(self, folder):
        p = self._absolute_output_path / folder
        os.makedirs(p, exist_ok=True)
        return p

    def initialize(self):
        # Sub-folders (relative to output_folder)
        self.optigan_input_folder = Path("optigan_inputs")
        self.optigan_output_folder = Path("optigan_outputs")
        self.optigan_csv_output_folder = self.optigan_output_folder / "csv_files"
        self.optigan_plots_folder = self.optigan_output_folder / "plots"

        # clear the input folder to make sure previous runs do not interfere
        delete_folder_contents(
            self.get_absolute_path_to_folder(self.optigan_input_folder)
        )
        self.device = get_torch_device(self.torch_device)

    def print_details_of_events(self):
        """
        Helper method to print event info.

        Not currently used by OptiGAN.
        DELETE in the future.
        """
        print(
            f"The length of extracted events details {len(self.extracted_events_details)}"
        )
        for event_id, detail in enumerate(self.extracted_events_details):
            gamma_pos = detail["gamma_position"]
            num_electrons = detail["electron_count"]
            num_optical_photons = detail["optical_photon_count"]
            print(
                f"Event ID: {event_id}, Gamma Position: {gamma_pos}, Number of Electrons: {num_electrons}, Number of Optical Photons: {num_optical_photons}"
            )
            print()

    def save_optigan_inputs(self):
        """
        Saves the required OptiGAN input as csv files.
        """

        for event_id, detail in enumerate(self.extracted_events_details):
            gamma_pos_x = detail["gamma_position"][0]
            gamma_pos_y = detail["gamma_position"][1]
            gamma_pos_z = detail["gamma_position"][2]
            num_optical_photons = detail["optical_photon_count"]

            data = {
                "gamma_pos_x": [gamma_pos_x],
                "gamma_pos_y": [gamma_pos_y],
                "gamma_pos_z": [gamma_pos_z],
                "num_optical_photons": [num_optical_photons],
            }

            pd.DataFrame(data).to_csv(
                self.get_absolute_path_to_folder(self.optigan_input_folder)
                / f"optigan_input_{event_id}.csv",
                index=False,
            )

            print(
                f"Event ID: {event_id}, Gamma Position: {gamma_pos_x}, {gamma_pos_y}, {gamma_pos_z}, "
                f"Number of Optical Photons: {num_optical_photons}"
            )
            print()

    def create_generator(self):
        """
        Loads the OptiGAN generator and prepares it for generating output.
        """

        input_dim, output_dim, hidden_dim, labels_len = self.gan_arguments.values()

        # Load the saved model checkpoint.
        checkpoint = torch.load(self.path_to_optigan_model, map_location=self.device)

        # Initialize the model.
        generator = WGANGenerator(input_dim, output_dim, hidden_dim, labels_len)

        # Load the state_dict into the model.
        generator.load_state_dict(checkpoint["generator_state_dict"])

        # Move the model to the appropriate device.
        generator.to(self.device)

        # Set the model to evaluation mode.
        generator.eval()

        return generator

    def generate_and_save_optigan_graphs(self):
        """
        Saves graphs in a different output folder.
        """
        csv_files = sorted(
            [
                file
                for file in os.listdir(
                    self.get_absolute_path_to_folder(self.optigan_csv_output_folder)
                )
                if file.endswith(".csv")
            ],
            key=extract_number,
        )

        for file_index, csv_file in enumerate(csv_files):
            df = pd.read_csv(
                self.get_absolute_path_to_folder(self.optigan_csv_output_folder)
                / csv_file
            )
            out_path = self.get_absolute_path_to_folder(
                self.optigan_plots_folder / f"event{file_index + 1}"
            )

            for column in df.columns:
                plt.figure(figsize=(10, 6))
                plt.hist(
                    df[column].dropna(), bins=30, color="blue", edgecolor="black"
                )  # Drop NaNs for cleaner plots
                plt.title(f"{column}: Distribution for event {file_index + 1}")
                plt.xlabel(column)
                plt.ylabel("Frequency")
                plt.tight_layout()

                # save the plot
                plt.savefig(out_path / f"{column}_event_{file_index + 1}.png")
                plt.close()

            print(
                f"The graphs for the output file {csv_file} are successfully created "
                f"in {out_path}"
            )

    def generate_and_save_optigan_output(self):
        """
        Generates and saves the optigan output as csv files.
        """

        # Sort and list CSV files.
        csv_files = sorted(
            [
                file
                for file in os.listdir(
                    self.get_absolute_path_to_folder(self.optigan_input_folder)
                )
                if file.endswith(".csv")
            ],
            key=extract_number,
        )

        output_path = self.get_absolute_path_to_folder(self.optigan_output_folder)
        # Clean the output folder.
        delete_folder_contents(output_path)
        print(f"The optigan output files will be saved at {output_path}")

        for file_index, file_name in enumerate(csv_files):
            # Prepare input file path and read csv.
            df = pd.read_csv(
                self.get_absolute_path_to_folder(self.optigan_input_folder) / file_name
            )

            # Extract and ensure total number of photons is a valid integer.
            total_number_of_photons = df["num_optical_photons"].values[0]
            print(f"Processing {file_name} with {total_number_of_photons} photons.")

            # Move the initial conditional values to the device.
            classX_single = torch.tensor(
                df["gamma_pos_x"].values, dtype=torch.float32
            ).to(self.device)
            classY_single = torch.tensor(
                df["gamma_pos_y"].values, dtype=torch.float32
            ).to(self.device)
            classZ_single = torch.tensor(
                df["gamma_pos_z"].values, dtype=torch.float32
            ).to(self.device)

            # Expand the conditional input vectors to match the total number of rows.
            classX = classX_single.expand(total_number_of_photons)
            classY = classY_single.expand(total_number_of_photons)
            classZ = classZ_single.expand(total_number_of_photons)

            # Create the random noise vector and combine conditions.
            noise = torch.randn(
                total_number_of_photons, self.gan_arguments["noise_dimension"]
            ).to(self.device)
            conditions = torch.stack([classX, classY, classZ], dim=1)

            # Concatenate noise and conditional input into one tensor
            generator_input = torch.cat((noise, conditions), dim=1)

            # Generate data using the model
            with torch.no_grad():
                generated_data = self.generator(generator_input)

            # Convert generated data to a DataFrame and save as a CSV file
            generated_data_np = generated_data.cpu().numpy()
            # generated_data_np = generated_data.to('cpu').detach().numpy()
            column_names = ["X", "Y", "dX", "dY", "dZ", "Ekine"]
            generated_df = pd.DataFrame(generated_data_np, columns=column_names)

            # Save the output CSV file
            optigan_output_csv_file_save_path = (
                self.get_absolute_path_to_folder(self.optigan_csv_output_folder)
                / f"optigan_output_{file_index + 1}.csv"
            )
            generated_df.to_csv(optigan_output_csv_file_save_path, index=False)
            print(f"Saved generated data to {optigan_output_csv_file_save_path}.")
        print()

    # Runs all the methods of OptiGAN
    def run_optigan(self, create_output_graphs):
        self.initialize()
        df_combined = self.get_dataframe_from_root_file()
        self.events = process_root_output_into_events(df_combined)
        self.extracted_events_details = self.extract_event_details()
        self.save_optigan_inputs()
        self.generator = self.create_generator()
        self.generate_and_save_optigan_output()

        if create_output_graphs:
            self.generate_and_save_optigan_graphs()

    def get_dataframe_from_root_file(self):
        """
        First step in the pipeline to extract OptiGAN input from the root output.
        Filters the raw root output into a format that is appropriate for OptiGAN.
        """

        with uproot.open(self.root_file_path) as file:
            root_tree = file["Phase"]

            # Save the particle co-ordinates and other information
            position_x = root_tree["Position_X"].array(library="np")
            position_y = root_tree["Position_Y"].array(library="np")
            position_z = root_tree["Position_Z"].array(library="np")
            particle_types = root_tree["ParticleName"].array(library="np")
            track_ids = root_tree["TrackID"].array(library="np")

        # Create a DataFrame from the final arrays.
        df = pd.DataFrame(
            {
                "TrackID": track_ids,
                "ParticleType": particle_types,
                "Position_X": position_x,
                "Position_Y": position_y,
                "Position_Z": position_z,
            }
        )

        df["OriginalIndex"] = df.index

        # Step 1: Filter out rows where ParticleType is "opticalphoton".
        opticalphoton_df = df[df["ParticleType"] == "opticalphoton"]

        # Step 2: Drop duplicates based on TrackID, keeping the first occurrence.
        opticalphoton_df_unique = opticalphoton_df.loc[
            ~opticalphoton_df.duplicated(subset="TrackID", keep="first")
        ]

        # Step 3: Filter out rows where ParticleType is NOT "opticalphoton" (e.g., "gamma").
        non_opticalphoton_df = df[df["ParticleType"] != "opticalphoton"]

        # Step 4: Concatenate both DataFrames to keep the structure intact.
        df_combined = pd.concat([opticalphoton_df_unique, non_opticalphoton_df])

        # Step 5: Sort the DataFrame based on index to restore the original order if needed.
        df_combined = df_combined.sort_index()

        return df_combined

    def extract_event_details(self):
        """
        Third step in the pipeline to extract event details from the root output.

        Extracts gamma, electron, and optical photon data from self.events.

        Note:
        - `electron_count` is unavailable because the phase space actor does not save
          this data.
        - Add a `Hits` actor in the simulation to obtain the required information
          (FUTURE FIX).
        """

        event_details = []
        for event_id, event in self.events.items():
            # dictionary format to store each event
            event_info = {
                "gamma_position": None,
                "electron_count": 0,
                "optical_photon_count": 0,
            }

            # loop through the particles in the event
            for particle in event:
                if particle["type"] == "gamma":
                    # save the position of the gamma particle
                    event_info["gamma_position"] = (
                        particle["x"],
                        particle["y"],
                        particle["z"],
                    )
                elif particle["type"] == "e-":
                    # increment the count of electrons
                    event_info["electron_count"] += 1
                elif particle["type"] == "opticalphoton":
                    event_info["optical_photon_count"] = particle[
                        "optical_photon_count"
                    ]

            if event_info["optical_photon_count"] != 0:
                event_details.append(event_info)

        return event_details



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.geometry.materials import MaterialDatabase
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, gate_folder="gate_test009_voxels", output_folder="test009"
    )

    # create the simulation
    sim = gate.Simulation()

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    mm = gate.g4_units.mm
    gcm3 = gate.g4_units.g_cm3
    is_ok = True

    # image
    image_volume = sim.add_volume("Image", "test1")
    image_volume.image = paths.data / "patient-4mm.mhd"
    image_volume.material = "G4_AIR"  # material used by default
    f1 = str(paths.gate_data / "Schneider2000MaterialsTable.txt")
    f2 = str(paths.gate_data / "Schneider2000DensitiesTable.txt")
    tol = 0.08 * gcm3
    vm, materials = gate.geometry.materials.HounsfieldUnit_to_material(sim, tol, f1, f2)
    image_volume.voxel_materials = vm

    # write mat and label-to-mat
    image_volume.write_material_database(paths.output / "test1.db")
    image_volume.write_label_to_material(paths.output / "test1.json")

    # read back
    sim.volume_manager.add_material_database(paths.output / "test1.db")
    vm = image_volume.voxel_materials.copy()
    image_volume.read_label_to_material(paths.output / "test1.json")
    print(f"Number of materials in the image: {len(vm)}")
    if image_volume.voxel_materials != vm:
        b = False
        utility.print_test(b, f"Error in the label to material {vm} ")
        utility.print_test(b, f"Error in the label to material {a}")
        is_ok = False

    # image
    image_volume = sim.add_volume("Image", "test2")
    image_volume.image = paths.data / "patient-4mm.mhd"
    image_volume.material = "G4_AIR"  # material used by default
    image_volume.voxel_materials = [
        [-2000, -900, "G4_AIR"],
        [-900, -100, "Lung"],
        [-100, 0, "G4_ADIPOSE_TISSUE_ICRP"],
        [0, 300, "G4_TISSUE_SOFT_ICRP"],
        [300, 800, "G4_B-100_BONE"],
        [800, 6000, "G4_BONE_COMPACT_ICRU"],
    ]
    image_volume.write_material_database(paths.output / "test2.db")
    image_volume.write_label_to_material(paths.output / "test2.json")

    # read back
    sim.volume_manager.add_material_database(paths.output / "test2.db")
    vm = image_volume.voxel_materials.copy()
    image_volume.read_label_to_material(paths.output / "test2.json")
    print(f"Number of materials in the image: {len(vm)}")
    if image_volume.voxel_materials != vm:
        b = False
        utility.print_test(b, f"Error in the label to material {vm} ")
        utility.print_test(b, f"Error in the label to material {a}")
        is_ok = False

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import opengate.contrib.linacs.elektasynergy as gate_linac
import gatetools.phsp as phsp
import matplotlib.pyplot as plt
from scipy.spatial.transform import Rotation

paths = utility.get_default_test_paths(
    __file__, "gate_test019_linac_phsp", output_folder="test019"
)


def init_test019(nt, version):
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    sim.number_of_threads = nt
    sim.random_seed = 123456789
    sim.output_dir = paths.output
    print(sim)

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    nm = gate.g4_units.nm
    Bq = gate.g4_units.Bq
    MeV = gate.g4_units.MeV

    #  adapt world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1.3 * m]

    # add a linac
    linac = gate_linac.add_linac(sim, "linac")
    linac.translation = [0, 0, 0 * m]

    # virtual plane for phase space
    plane = sim.add_volume("Tubs", "phase_space_plane")
    plane.mother = linac.name
    plane.material = "G4_AIR"
    plane.rmin = 0
    plane.rmax = 70 * mm
    plane.dz = 1 * nm  # half height
    plane.translation = [0, 0, linac.size[2] / 2 - 300 * mm]
    plane.color = [1, 0, 0, 1]  # red

    # e- source
    source = sim.add_source("GenericSource", "Default")
    source.particle = "e-"
    source.attached_to = f"{linac.name}_target"
    source.energy.type = "gauss"
    source.energy.mono = 6.7 * MeV
    source.energy.sigma_gauss = 0.077 * MeV
    source.position.type = "disc"
    source.position.radius = 2 * mm  # FIXME not really similar to GATE need sigma etc
    source.position.translation = [0, 0, 0.5 * mm - 1 * nm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, -1]
    source.activity = 5000 * Bq / sim.number_of_threads

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "Stats")
    s.track_types_flag = True

    # PhaseSpace Actor
    ta2 = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    ta2.attached_to = plane.name
    ta2.attributes = [
        "KineticEnergy",
        "Weight",
        "PostPosition",
        "PrePosition",
        "PrePositionLocal",
        "ParticleName",
        "PreDirection",
        "PreDirectionLocal",
        "PostDirection",
        "TimeFromBeginOfEvent",
        "GlobalTime",
        "LocalTime",
        "EventPosition",
        "PDGCode",
    ]
    ta2.output_filename = f"test019_hits_{version}.root"
    ta2.debug = False
    ta2.steps_to_store = "exiting"
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "gamma"
    ta2.filters.append(f)

    # phys
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 1 * mm
    sim.physics_manager.global_production_cuts.electron = 1 * mm
    sim.physics_manager.global_production_cuts.positron = 1 * mm

    return sim


def run_test019(sim):
    # splitting
    linac = sim.volume_manager.get_volume("linac")
    region_linac = sim.physics_manager.add_region(name=f"{linac.name}_region")
    region_linac.associate_volume(linac)
    # FIXME: should be a user info in Region
    s = f"/process/em/setSecBiasing eBrem {region_linac.name} 100 100 MeV"
    print(s)
    sim.g4_commands_after_init.append(s)

    # start simulation
    sim.run()

    # print results
    stats = sim.get_actor("Stats")
    print(stats)

    h = sim.get_actor("PhaseSpace")
    print(h)

    # FIXME: missing features:
    # - missing several branch names in PhaseSpaceActor
    # - no local/global for position
    # - no policy options (all track single etc)

    # check stats
    print()
    stats_ref = utility.read_stat_file(paths.gate_output / "output-writePhS-stat.txt")
    print(f"Number of runs was {stats.counts.runs}. Set to 1 before comparison")
    stats.counts.runs = 1
    is_ok = utility.assert_stats(stats, stats_ref, 0.2)

    # compare the phsp tree
    print()
    fn1 = paths.gate_output / "output-PhS-g.root"
    fn2 = h.get_output_path()
    print("Reference gate tree : ", fn1)
    print("Checked Tree : ", fn2)
    data_ref, keys_ref, m_ref = phsp.load(fn1)
    data, keys, m = phsp.load(fn2)
    # find the good key's names
    keys1, keys2, scalings, tols = utility.get_keys_correspondence(keys_ref)
    keys1.remove("Z")
    keys2.remove("PrePosition_Z")
    # Do not check some keys
    tols[keys1.index("Weight")] = 0.001
    tols[keys1.index("Ekine")] = 0.1
    tols[keys1.index("Y")] = 2.0
    tols[keys1.index("X")] = 3.0
    # tols[keys1.index("Z")] = 0.2
    # perform the test
    is_ok = (
        utility.compare_trees(
            data_ref, keys_ref, data, keys, keys1, keys2, tols, scalings, scalings, True
        )
        and is_ok
    )

    # figure
    plt.suptitle(f"Values: {len(data_ref)} vs {len(data)}")
    # plt.show()
    fn = paths.output / "test019.png"
    plt.savefig(fn)
    print(f"Figure in {fn}")

    # this is the end, my friend
    utility.test_ok(is_ok)


def create_simu_test019_phsp_source(sim, version):
    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    # sim.running_verbose_level = gate.logger.EVENT
    sim.number_of_threads = 1
    sim.random_seed = 987654321
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    nm = gate.g4_units.nm

    #  adapt world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # virtual plane for phase space
    plane = sim.add_volume("Tubs", "phase_space_plane")
    plane.material = "G4_AIR"
    plane.rmin = 0
    plane.rmax = 70 * mm
    plane.dz = 1 * nm  # half height
    plane.rotation = Rotation.from_euler("xy", [180, 30], degrees=True).as_matrix()
    plane.translation = [-5 * mm, 20 * mm, 33 * mm]
    plane.color = [1, 0, 0, 1]  # red

    plane2 = sim.add_volume("Tubs", "phase_space_plane2")
    plane2.material = "G4_AIR"
    plane2.rmin = 0
    plane2.rmax = 70 * mm
    plane2.dz = 1 * nm  # half height
    plane2.translation = [0 * mm, 0 * mm, -300.001 * mm]
    plane2.color = [1, 0, 0, 1]  # red

    # phsp source
    source = sim.add_source("PhaseSpaceSource", "phsp_source_local")
    source.attached_to = plane.name
    source.phsp_file = paths.output_ref / "test019_hits.root"
    source.position_key = "PrePositionLocal"
    source.direction_key = "PreDirectionLocal"
    source.global_flag = False
    source.particle = "gamma"
    source.particle = ""
    source.PDGCode_key = "PDGCode"
    source.n = 20000 / sim.number_of_threads
    source.batch_size = source.n
    source.verbose = False
    source.verbose_batch = True

    # phsp source
    source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    source.attached_to = sim.world.name
    source.phsp_file = paths.output_ref / "test019_hits.root"
    source.position_key = "PrePosition"
    source.direction_key = "PreDirection"
    source.global_flag = True
    source.particle = None
    source.PDGCode_key = "PDGCode"
    source.n = 20000 / sim.number_of_threads
    source.batch_size = source.n
    source.verbose = False
    source.verbose_batch = True

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "Stats")
    s.track_types_flag = True

    # PhaseSpace Actor
    ta1 = sim.add_actor("PhaseSpaceActor", "PhaseSpace1")
    ta1.attached_to = plane.name
    ta1.attributes = [
        "KineticEnergy",
        "Weight",
        "PostPosition",
        "PrePosition",
        "PrePositionLocal",
        "PostPositionLocal",
        "ParticleName",
        "PreDirection",
        "PreDirectionLocal",
        "PostDirection",
        "TimeFromBeginOfEvent",
        "GlobalTime",
        "LocalTime",
        "EventPosition",
    ]
    ta1.output_filename = "test019_hits_phsp_source_local{version}.root"
    ta1.debug = False
    ta1.steps_to_store = "exiting"
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "gamma"
    ta1.filters.append(f)

    # PhaseSpace Actor
    ta2 = sim.add_actor("PhaseSpaceActor", "PhaseSpace2")
    ta2.attached_to = plane2.name
    ta2.attributes = [
        "KineticEnergy",
        "Weight",
        "PostPosition",
        "PrePosition",
        "PrePositionLocal",
        "ParticleName",
        "PreDirection",
        "PreDirectionLocal",
        "PostDirection",
        "TimeFromBeginOfEvent",
        "GlobalTime",
        "LocalTime",
        "EventPosition",
    ]
    ta2.output_filename = "test019_hits_phsp_source_global{version}.root"
    ta2.debug = False
    f = sim.add_filter("ParticleFilter", "f2")
    f.particle = "gamma"
    ta2.filters.append(f)

    # phys
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 1 * mm
    sim.physics_manager.global_production_cuts.electron = 1 * mm
    sim.physics_manager.global_production_cuts.positron = 1 * mm


def analyse_test019_phsp_source(sim):
    # print results
    stats = sim.get_actor("Stats")
    print(stats)

    # print source phsp info
    s1 = sim.source_manager.get_source("phsp_source_local")
    print(f"Source local :  {s1.num_entries} elements, {s1.cycle_count} cycle")
    s2 = sim.source_manager.get_source("phsp_source_global")
    print(f"Source global : {s2.num_entries} elements, {s2.cycle_count} cycle")

    # --------------------------------------------------------------
    # Test LOCAL position
    print()
    print("Test LOCAL position")
    fn1 = paths.output_ref / "test019_hits.root"
    ta1 = sim.get_actor("PhaseSpace1")
    fn2 = ta1.get_output_path()
    print("Reference gate tree : ", fn1)
    print("Checked Tree : ", fn2)
    data, keys, _ = phsp.load(fn2, "PhaseSpace1")
    data_ref, keys_ref, _ = phsp.load(fn1, nmax=len(data))

    keys1 = [
        "PrePositionLocal_X",
        "PrePositionLocal_Y",
        "PrePositionLocal_Z",
        "KineticEnergy",
        "Weight",
    ]
    keys2 = keys1
    tols = [0.01] * len(keys1)
    scalings = [1.0] * len(keys1)
    is_ok = utility.compare_trees(
        data_ref, keys_ref, data, keys, keys1, keys2, tols, scalings, scalings, True
    )

    # figure
    plt.suptitle(f"Values: {len(data_ref)} vs {len(data)}")
    # plt.show()
    fn = paths.output / "test019_source_local.png"
    plt.savefig(fn)
    print(f"Figure in {fn}")

    # --------------------------------------------------------------
    # Test GLOBAL position
    print()
    print("Test GLOBAL position")
    fn1 = paths.output_ref / "test019_hits.root"
    ta2 = sim.get_actor("PhaseSpace2")
    fn2 = ta2.get_output_path()
    print("Reference gate tree : ", fn1)
    print("Checked Tree : ", fn2)
    data, keys, _ = phsp.load(fn2, "PhaseSpace2")
    data_ref, keys_ref, _ = phsp.load(fn1, nmax=len(data))

    keys1 = [
        "PrePosition_X",
        "PrePosition_Y",
        "PrePosition_Z",
        "KineticEnergy",
        "Weight",
    ]
    keys2 = keys1
    tols = [0.01] * len(keys1)
    scalings = [1.0] * len(keys1)
    is_ok = (
        utility.compare_trees(
            data_ref, keys_ref, data, keys, keys1, keys2, tols, scalings, scalings, True
        )
        and is_ok
    )

    # figure
    plt.suptitle(f"Values: {len(data_ref)} vs {len(data)}")
    # plt.show()
    fn = paths.output / "test019_source_global.png"
    plt.savefig(fn)
    print(f"Figure in {fn}")

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect.siemens_intevo as intevo
import opengate.contrib.spect.ge_discovery_nm670 as nm670
from opengate.geometry.volume_info import store_volumes_info


def vol_info(sim, filename):
    store_volumes_info(sim, filename)
    print(f"opengate_plot_volume_info {filename} --hl arf")


def build_test_simu(
    paths, spect_model, colli_type, rotation, radius, crystal_size=None
):
    sim = gate.Simulation()
    sim.verbose_level = gate.logger.NONE
    sim.output_dir = paths.output
    sim.visu_type = "qt"
    sim.user_hook_after_init = vol_info
    if crystal_size is None:
        fn = f"vol_info_{spect_model}_{colli_type}_{rotation}.json"
    else:
        fn = f"vol_info_{spect_model}_{crystal_size.replace('/', '-')}_{colli_type}_{rotation}.json"
    sim.user_hook_after_init_arg = sim.output_dir / fn

    m = None
    head = None
    if spect_model == "intevo":
        m = intevo
        head, colli, crystal = m.add_spect_head(
            sim, "spect", collimator_type=colli_type, debug=sim.visu == True
        )

    if spect_model == "nm670":
        m = nm670
        head, colli, crystal = m.add_spect_head(
            sim,
            "spect",
            collimator_type=colli_type,
            debug=sim.visu == True,
            crystal_size=crystal_size,
        )

    # sim.visu = True # debug

    # add the arf plane
    arf_plane = m.add_detection_plane_for_arf(sim, "arf", colli_type)

    # rotate the gantry of the real spect
    m.rotate_gantry(head, radius=radius, start_angle_deg=rotation)

    # rotate the gantry of the arf plane
    m.rotate_gantry(arf_plane, radius=radius, start_angle_deg=rotation)

    return sim, fn



import multiprocessing
import queue
from .exception import fatal
import os


# define a thin wrapper function to handle the queue
def target_func(q, f, *args, **kwargs):
    q.put(f(*args, **kwargs))


def dispatch_to_subprocess(func, *args, **kwargs):
    if os.name == "nt":
        # Windows: 'spawn' is required
        start_method = "spawn"
    else:
        # Unix (Linux or macOS)
        # On Unix, 'fork' is faster than 'spawn' but require Manager().Queue()
        start_method = "fork"
    try:
        multiprocessing.set_start_method(start_method)
    except RuntimeError:
        pass

    # Queue is faster than Manager().Queue() but fails with fork
    # q = multiprocessing.Queue()
    q = multiprocessing.Manager().Queue()
    p = multiprocessing.Process(
        target=target_func, args=(q, func, *args), kwargs=kwargs
    )
    p.start()
    p.join()

    try:
        return q.get(block=False)
    except queue.Empty:
        fatal("The queue is empty. The spawned process probably died.")
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from matplotlib import pyplot as plt
import opengate as gate
from opengate import g4_units
from opengate.tests import utility
from opengate.tests.src.test081_tle_helpers import (
    add_waterbox,
    voxelize_waterbox,
    add_source,
    add_iso_source,
    plot_pdd,
    compare_pdd,
)
from opengate.tests.utility import get_image_1d_profile

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test081_tle")

    # create the simulation
    sim = gate.Simulation()

    # main options
    # sim.visu = True
    sim.visu_type = "qt"
    sim.random_seed = 321654
    sim.output_dir = paths.output
    sim.progress_bar = True
    sim.number_of_threads = 1

    # units
    gcm3 = gate.g4_units.g / gate.g4_units.cm3
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    keV = gate.g4_units.keV
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # insert voxelized waterbox
    fn = paths.data / "test081_tle"
    box = sim.add_volume("Image", "box")
    box.image = f"{fn}/random_HU_img.mhd"
    box.mother = "world"
    box.material = "G4_AIR"  # material used by default
    f1 = paths.data / "Schneider2000MaterialsTable.txt"
    f2 = paths.data / "Schneider2000DensitiesTable.txt"
    tol = 0.05 * gcm3
    box.voxel_materials, materials = gate.geometry.materials.HounsfieldUnit_to_material(
        sim, tol, f1, f2
    )
    box.color = [1, 0, 1, 1]
    box_size = [180, 180, 180]
    # waterbox_size = [30 * cm, 30 * cm, 20 * cm]

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.global_production_cuts.all = 1 * mm
    s = f"/process/eLoss/CSDARange true"
    sim.g4_commands_before_init.append(s)

    # default source for tests
    source = add_iso_source(sim, n=4e5)

    # add tle dose actor
    tle_dose_actor = sim.add_actor("TLEDoseActor", "tle_dose_actor")
    tle_dose_actor.output_filename = "test081_vox_tle_rd_HU.mhd"
    tle_dose_actor.attached_to = box
    tle_dose_actor.dose_uncertainty.active = True
    tle_dose_actor.dose.active = True
    tle_dose_actor.size = [20, 20, 20]
    tle_dose_actor.spacing = [x / y for x, y in zip(box_size, tle_dose_actor.size)]
    tle_dose_actor.density.active = True
    tle_dose_actor.tle_threshold_type = "average range"
    tle_dose_actor.tle_threshold = 3 * mm
    # tle_dose_actor.tle_threshold_type = "energy"
    # tle_dose_actor.tle_threshold = 1*MeV
    tle_dose_actor.score_in = "material"  # only 'material' is allowed
    print(f"TLE Dose actor pixels : {tle_dose_actor.size}")
    print(f"TLE Dose actor spacing : {tle_dose_actor.spacing} mm")
    print(f"TLE Dose actor size : {box_size} mm")

    # add conventional dose actor
    dose_actor = sim.add_actor("DoseActor", "dose_actor")
    dose_actor.output_filename = "test081_vox_rd_HU.mhd"
    dose_actor.attached_to = box
    dose_actor.dose_uncertainty.active = True
    dose_actor.dose.active = True
    dose_actor.size = [20, 20, 20]
    dose_actor.spacing = [x / y for x, y in zip(box_size, dose_actor.size)]
    dose_actor.density.active = True
    print(f"Dose actor pixels : {dose_actor.size}")
    print(f"Dose actor spacing : {dose_actor.spacing} mm")
    print(f"Dose actor size : {box_size} mm")

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    ax, plt = plot_pdd(dose_actor, tle_dose_actor)
    f1 = dose_actor.edep.get_output_path()
    f2 = tle_dose_actor.edep.get_output_path()
    is_ok = compare_pdd(f1, f2, dose_actor.spacing[2], ax[0], tol=0.1)

    f1 = dose_actor.dose.get_output_path()
    f2 = tle_dose_actor.dose.get_output_path()
    is_ok = compare_pdd(f1, f2, dose_actor.spacing[2], ax[1], tol=0.1) and is_ok

    # output
    f = paths.output / f"pdd_vox_rd_HU.png"
    plt.savefig(f)
    print(f"PDD image saved in {f}")

    # check density
    utility.assert_images(
        dose_actor.density.get_output_path(),
        tle_dose_actor.density.get_output_path(),
        tolerance=0.001,
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import uproot
import sys


def test022_half_life(n_threads=1):
    paths = utility.get_default_test_paths(__file__, output_folder="test022")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.number_of_threads = n_threads
    sim.random_seed = 12344321
    sim.output_dir = paths.output
    print(sim)

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    nm = gate.g4_units.nm
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.s

    # set the world size like in the Gate macro
    sim.world.size = [1 * m, 1 * m, 2 * m]

    # waterbox (not really used here)
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [10 * cm, 10 * cm, 10 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 0 * cm]
    waterbox.material = "G4_AIR"

    # detector
    detector = sim.add_volume("Box", "detector")
    detector.size = [80 * cm, 80 * cm, 1 * nm]
    detector.translation = [0, 0, 30 * cm]
    detector.material = "G4_BGO"
    detector.color = [1, 0, 0, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"
    sim.physics_manager.enable_decay = True
    sim.physics_manager.global_production_cuts.all = (
        1 * mm
    )  # all means: protons, electrons, positrons, gammas

    # source #1
    source1 = sim.add_source("GenericSource", "source1")
    source1.particle = "gamma"
    source1.energy.mono = 100 * keV
    source1.position.type = "disc"
    source1.position.radius = 2 * cm
    source1.position.translation = [0, 0, -10 * cm]
    source1.direction.type = "focused"
    source1.direction.focus_point = [0, 0, 0]
    source1.activity = 10000 * Bq / sim.number_of_threads
    source1.half_life = 2 * sec

    # source #2
    source2 = sim.add_source("GenericSource", "source2")
    source2.particle = "gamma"
    source2.energy.mono = 200 * keV
    source2.position.type = "disc"
    source2.position.radius = 2 * cm
    source2.position.translation = [0, 0, -10 * cm]
    source2.direction.type = "focused"
    source2.direction.focus_point = [0, 0, 0]
    source2.activity = 10000 * Bq / sim.number_of_threads
    # source2.n = 50

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # hit actor
    ta = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    ta.attached_to = "detector"
    ta.attributes = ["KineticEnergy", "GlobalTime"]
    ta.output_filename = f"test022_half_lifev_{n_threads}.root"

    # timing
    sim.run_timing_intervals = [
        [1 * sec, 10 * sec],
        [15 * sec, 20 * sec],
    ]  # "hole" in the timeline

    # start simulation
    sim.run(start_new_process=True)

    # get result
    print(stats)

    # read phsp
    with uproot.open(ta.get_output_path()) as root:
        branch = root["PhaseSpace"]["GlobalTime"]
        time = branch.array(library="numpy") / sec
        branch = root["PhaseSpace"]["KineticEnergy"]
        E = branch.array(library="numpy")

    # consider time of arrival for both sources
    time1 = time[E < 110 * keV]
    time2 = time[E > 110 * keV]

    # fit for half life
    start_time = sim.run_timing_intervals[0][0] / sec
    end_time = sim.run_timing_intervals[0][1] / sec
    hl, xx, yy = utility.fit_exponential_decay(time1, start_time, end_time)
    # compare with source half_life (convert in sec)
    tol = 0.05
    hl_ref = source1.half_life / sec
    diff = abs(hl - hl_ref) / hl_ref
    is_ok = b = diff < tol
    diff *= 100
    utility.print_test(b, f"Half life {hl_ref:.2f} sec vs {hl:.2f} sec : {diff:.2f}% ")

    # check second source
    m = len(time2)
    start_time2 = sim.run_timing_intervals[1][0] / sec
    end_time2 = sim.run_timing_intervals[1][1] / sec
    # number of elements is around activity times the duration (per thread)
    m_ref = (
        source2.activity
        / Bq
        * (end_time - start_time + end_time2 - start_time2)
        * sim.number_of_threads
    )
    diff = abs(m - m_ref) / m_ref
    b = diff < tol
    diff *= 100
    utility.print_test(b, f"Events for source #2:  {m_ref} vs {m} -> {diff:.2f}% ")
    is_ok = is_ok and b

    # check thread
    b = sim.number_of_threads * len(sim.run_timing_intervals) == stats.counts.runs
    utility.print_test(b, f"Number of run: {stats.counts.runs}")

    is_ok = is_ok and b

    # plot debug
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(5, 5))
    a = ax
    a.hist(
        time1,
        bins=100,
        label="decay source",
        histtype="stepfilled",
        alpha=0.5,
        density=True,
    )
    a.hist(
        time2,
        bins=100,
        label="constant source",
        histtype="stepfilled",
        alpha=0.5,
        density=True,
    )
    a.plot(xx, yy, label="fit half-life {:.2f} sec".format(hl))
    a.legend()
    a.set_xlabel("time (s)")
    a.set_ylabel("detected photon")
    # plt.show()

    fn = paths.output / "test022_half_life_fit.png"
    print("Figure in ", fn)
    plt.savefig(fn)

    utility.test_ok(is_ok)


def main():
    test022_half_life(n_threads=1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Context: By default, simulation of e+ source does not result in the expected acolin. of
PET imaging. To enable this, one need to set ionisation of the material where the
annihilations will occur to 5.0 eV. The test is in two part:
1) If nothing is done, more annihilations should be colinear
2) If the mean energy per ion pair is set to 5.0 eV, the amplitude of acolinearity
should follow a Rayleight distribution with a scale of 0.21 deg., which corresponds to
the acolin deviation following a 2D Gaussian with a FWHM of 0.5 deg.

Here, the material, G4_WATER is already known of Geant4, so one only need to set its
ionisation correctly.
"""

from test079_acollin_helpers import *
import opengate.tests.utility as tu
import matplotlib.pyplot as plt


#########################################################################################
# Simulations configuration that may be relevant to change
#########################################################################################
# Mean energy of Ion Pair to use. 5.0 eV should produce the expected 0.5 deg FWHM in PET
# imaging
mean_energy = 5.0 * eV
# Key added to output to make sure that multi-threading the tests does not backfire
test_key = "p4"


#########################################################################################
# Main : We use this to launch the test
#########################################################################################
if __name__ == "__main__":
    paths = tu.get_default_test_paths(__file__, output_folder="test079")

    # Define core of the simulation, including physics
    sim = setup_simulation_engine(paths)
    sim.random_seed = 12345654
    sim.progress_bar = True

    # add a waterbox
    wb = sim.add_volume("Box", "waterbox")
    wb.size = [50 * cm, 50 * cm, 50 * cm]
    wb.material = "G4_WATER"

    # test mat properties
    mat = sim.volume_manager.find_or_build_material(wb.material)
    ionisation = mat.GetIonisation()
    print(
        f"material {wb.material} mean excitation energy is {ionisation.GetMeanExcitationEnergy() / eV} eV"
    )
    print(
        f"material {wb.material} mean energy per ion pair is {ionisation.GetMeanEnergyPerIonPair() / eV} eV"
    )

    # set the source
    source = sim.add_source("GenericSource", "beta+_source")
    source.particle = "e+"
    source.energy.type = "F18"
    source.activity = 10000 * Bq
    source.direction.type = "iso"

    # add phase actor
    phsp = setup_actor(sim, "phsp", wb.name)
    phsp.output_filename = paths.output / f"annihilation_photons_{test_key}.root"

    # go
    sim.run(start_new_process=True)

    # redo test changing the MeanEnergyPerIonPair
    root_filename = phsp.output_filename
    phsp.output_filename = (
        paths.output / f"annihilation_photons_with_mepip_{test_key}.root"
    )
    ionisation.SetMeanEnergyPerIonPair(mean_energy)
    print(f"set MeanEnergyPerIonPair to {ionisation.GetMeanEnergyPerIonPair() / eV} eV")

    # go
    sim.run()

    # test: no mean energy, should be mostly colinear
    gamma_pairs = read_gamma_pairs(root_filename)
    acollinearity_angles = compute_acollinearity_angles(gamma_pairs)

    colin_median = plot_colin_case(acollinearity_angles)

    # test: with mean energy, acolinearity amplitude should have a Rayleigh distribution
    gamma_pairs = read_gamma_pairs(phsp.output_filename)
    acollinearity_angles = compute_acollinearity_angles(gamma_pairs)

    acolin_scale = plot_acolin_case_mepip(mean_energy, acollinearity_angles)

    f = paths.output / f"acollinearity_angles_{test_key}.png"
    plt.savefig(f)
    print(f"Plot was saved in {f}")

    # final
    # No acolin
    is_ok_p1 = colin_median < 0.01
    # Basic acolin
    is_ok_p2 = np.isclose(acolin_scale * 2.355, 0.5, atol=0.2)

    tu.test_ok(is_ok_p1 and is_ok_p2)



import pathlib
from opengate.geometry.volumes import unite_volumes
from opengate.geometry.volumes import RepeatParametrisedVolume, BoxVolume
from opengate.actors.digitizers import *
from opengate.managers import Simulation
from opengate.utility import g4_units
from opengate.contrib.spect.spect_helpers import (
    get_volume_position_in_head,
    get_default_energy_windows,
)
from opengate.geometry.utility import get_transform_orbiting
from opengate.contrib.spect.spect_helpers import (
    get_mu_from_xraylib,
    calculate_theta_max_angle,
)
from box import Box
import json


def get_geometrical_parameters_filename():
    filename = (
        pathlib.Path(__file__).parent
        / "spect_siemens_intevo_geometrical_parameters.json"
    )
    return pathlib.Path(filename)


geometrical_parameters = None


def get_geometrical_parameters():
    global geometrical_parameters
    if geometrical_parameters is None:
        filename = get_geometrical_parameters_filename()
        if not filename.exists():
            print(f'update geometrical parameters to "{filename}"')
            update_geometrical_parameters(store_to_file=True)
        # print(f'Loading geometrical parameters from "{filename}"')
        with open(filename) as json_file:
            geometrical_parameters = json.load(json_file)
        geometrical_parameters = Box(geometrical_parameters)
    return geometrical_parameters


def update_geometrical_parameters(store_to_file=False):
    p = Box()
    p.collimators = ["lehr", "melp", "he"]
    # for all colli
    nm = g4_units.nm
    for c in p.collimators:
        s = Simulation()
        spect, colli, crystal = add_spect_head(s, "spect", c, debug=True)
        pos = get_volume_position_in_head(s, "spect", f"collimator", "min", axis=0)
        y = get_volume_position_in_head(s, "spect", "crystal", "center", axis=0)
        psd = get_volume_position_in_head(s, "spect", "shielding_front", "min", axis=0)
        p[c] = Box()
        # distance from box boundary to collimator
        p[c].collimator_position = pos
        # distance to the center of the box head
        p[c].half_box_size = spect.size[0] / 2.0 + 1 * nm
        # distance from box boundary to the crystal center (for arf)
        p[c].crystal_distance = p[c].half_box_size + y
        # distance from box boundary to the shielding front
        p[c].psd = p[c].half_box_size + psd

        # collimator holes
        hole = s.volume_manager.get_volume(f"spect_collimator_hole1")
        p[c].hole_diameter = hole.radius * 2
        p[c].collimator_length = hole.height
        holep = s.volume_manager.get_volume(f"spect_collimator_hole1_param")
        tr = holep.translation
        p[c].septa_thickness = tr[1] - p[c].hole_diameter

    if store_to_file:
        filename = get_geometrical_parameters_filename()
        with open(filename, "w") as json_file:
            json.dump(p, json_file, indent=4)

    return p


def add_spect_head(sim, name="spect", collimator_type="lehr", debug=False):
    """
    Collimator:
    - False or "none" : no collimator
    - LEHR  : 1.11 mm holes
    - MELP  : 2.94 mm holes
    - HE    : 4 mm holes

    Collimator LEHR: Low Energy High Resolution    (for Tc99m)
    Collimator MELP: Medium Energy Low Penetration (for In111, Lu177)
    Collimator HE:   High Energy General Purpose   (for I131)

    By default, a translation is set such as the shielding_front is at position 0.
    """
    f = pathlib.Path(__file__).parent.resolve()
    fdb = f"{f}/spect_siemens_intevo_materials.db"
    if fdb not in sim.volume_manager.material_database.filenames:
        sim.volume_manager.add_material_database(fdb)

    # check overlap
    sim.check_volumes_overlap = debug

    # main box
    head = add_head_box(sim, name)

    # shielding (important)
    add_shielding(sim, head, collimator_type)

    # collimator
    colli = add_collimator(sim, head, collimator_type, debug)

    # crystal
    crystal = add_crystal(sim, head)

    # other elements
    back_comp = add_back_compartment(sim, head)
    light_guide = add_light_guide(sim, back_comp)
    # pmt = add_PMT_array(sim, head)
    # elec = add_electronics(sim, head)

    # default head translation to set the shielding_front at position 0
    # (this translation is considered in rotate_gantry)
    p = get_geometrical_parameters()
    r = p[collimator_type].half_box_size - p[collimator_type].psd
    head.translation = [0, r, 0]

    return head, colli, crystal


def add_head_box(sim, name):
    mm = g4_units.mm
    # bounding box
    head = sim.add_volume("Box", name)
    head.material = "G4_AIR"
    head.size = [260.0448 * mm, 685 * mm, 539 * mm]
    white = [1, 1, 1, 1]
    head.color = white
    return head


def add_shielding(sim, head, collimator_type):
    if collimator_type == "lehr" or collimator_type == "melp":
        return add_shielding_lehr_melp(sim, head)
    if collimator_type == "he":
        return add_shielding_he(sim, head)


def add_shielding_lehr_melp(sim, head):
    mm = g4_units.mm
    # shielding side thickness
    side_thickness = 12.7 * mm
    front_depth = 60 * mm
    side_depth = 215.0448 * mm - front_depth
    name = head.name

    # shielding
    back_shield = BoxVolume(name=f"{name}_back_shield")
    back_shield.size = [17.1 * mm, head.size[1], head.size[2]]

    # side shield Y pos
    side_shield_y_pos = BoxVolume(name=f"{name}_side_shield_y_pos")
    side_shield_y_pos.size = [side_depth, side_thickness, head.size[2]]

    # side shield Z pos
    side_shield_z_pos = BoxVolume(name=f"{name}_side_shield_z_pos")
    # gate.geometry.utility.copy_volume_user_info(side_shield_y_pos, side_shield_z_pos)
    side_shield_z_pos.size = [side_depth, head.size[1], side_thickness]

    # first volume from the 5 solids
    dx = -98.9724 * mm
    ddx = dx + front_depth / 2
    a = unite_volumes(back_shield, side_shield_y_pos, [ddx, 336.15 * mm, 0])
    a = unite_volumes(a, side_shield_y_pos, [ddx, -336.15 * mm, 0])
    a = unite_volumes(a, side_shield_z_pos, [ddx, 0, 263.15 * mm])
    shield = unite_volumes(
        a,
        side_shield_z_pos,
        [ddx, 0, -263.15 * mm],
        new_name=f"{name}_shielding_back_side",
    )
    sim.add_volume(shield)
    shield.mother = head.name
    shield.translation = [-dx, 0, 0]
    gray = [0.5, 0.5, 0.5, 1]
    shield.color = gray
    shield.material = "Lead"

    # front shields
    front_shield_y_pos = BoxVolume(name=f"{name}_front_shield_y_pos")
    front_shield_y_pos.size = [front_depth, 76 * mm, 510.4 * mm]
    front_shield_z_pos = BoxVolume(name=f"{name}_front_shield_z_pos")
    front_shield_z_pos.size = [front_depth, 676.7 * mm, 76 * mm]

    # second volume from the 4 solids
    a = front_shield_y_pos
    t = 304.5 * mm
    a = unite_volumes(a, front_shield_y_pos, [0, -2 * t, 0])
    a = unite_volumes(a, front_shield_z_pos, [0, -t, 231.5 * mm])
    shield = unite_volumes(
        a, front_shield_z_pos, [0, -t, -231.5 * mm], new_name=f"{name}_shielding_front"
    )
    sim.add_volume(shield)
    shield.mother = head.name
    shield.translation = [-87.0776 * mm, t, 0]
    shield.color = gray
    shield.material = "Lead"

    return shield


def add_shielding_he(sim, head):
    mm = g4_units.mm
    # shielding side thickness
    side_thickness = 12.7 * mm
    front_depth = 60 * mm
    side_depth = 215.0448 * mm - front_depth
    name = head.name

    # shielding
    back_shield = BoxVolume(name=f"{name}_back_shield")
    back_shield.size = [17.1 * mm, head.size[1], head.size[2]]

    # side shield Y pos
    side_shield_y_pos = BoxVolume(name=f"{name}_side_shield_y_pos")
    side_shield_y_pos.size = [side_depth, side_thickness, head.size[2]]

    # side shield Z pos
    side_shield_z_pos = BoxVolume(name=f"{name}_side_shield_z_pos")
    # gate.geometry.utility.copy_volume_user_info(side_shield_y_pos, side_shield_z_pos)
    side_shield_z_pos.size = [side_depth, head.size[1], side_thickness]

    # first volume from the 5 solids
    dx = -98.9724 * mm
    ddx = dx + front_depth / 2
    a = unite_volumes(back_shield, side_shield_y_pos, [ddx, 336.15 * mm, 0])
    a = unite_volumes(a, side_shield_y_pos, [ddx, -336.15 * mm, 0])
    a = unite_volumes(a, side_shield_z_pos, [ddx, 0, 263.15 * mm])
    shield = unite_volumes(
        a,
        side_shield_z_pos,
        [ddx, 0, -263.15 * mm],
        new_name=f"{name}_shielding_back_side",
    )
    sim.add_volume(shield)
    shield.mother = head.name
    shield.translation = [-dx, 0, 0]
    gray = [0.5, 0.5, 0.5, 1]
    shield.color = gray
    shield.material = "Lead"

    # front shields
    ofy = 35 * mm
    ofz = 30 * mm
    front_shield_y_pos = BoxVolume(name=f"{name}_front_shield_y_pos")
    front_shield_y_pos.size = [front_depth, 76 * mm - ofy, 510.4 * mm]
    front_shield_z_pos = BoxVolume(name=f"{name}_front_shield_z_pos")
    front_shield_z_pos.size = [front_depth, 676.7 * mm, 76 * mm - ofz]

    # second volume from the 4 solids
    a = front_shield_y_pos
    ty = 304.5 * mm + ofy / 2
    tz = 231.5 * mm + ofz / 2
    a = unite_volumes(a, front_shield_y_pos, [0, -2 * ty, 0])
    a = unite_volumes(a, front_shield_z_pos, [0, -ty, tz])
    shield = unite_volumes(
        a, front_shield_z_pos, [0, -ty, -tz], new_name=f"{name}_shielding_front"
    )
    sim.add_volume(shield)
    shield.mother = head.name
    shield.translation = [-87.0776 * mm, ty, 0]
    blue = [0.5, 0.5, 1, 0.8]
    shield.color = blue
    shield.material = "Lead"

    return shield


def add_collimator(sim, head, collimator_type, debug):
    if (
        collimator_type is False
        or collimator_type is None
        or collimator_type == ""
        or collimator_type == "none"
    ):
        return add_collimator_empty(sim, head)
    collimator_type = collimator_type.lower()
    if collimator_type == "lehr":
        return add_collimator_lehr(sim, head, debug)
    if collimator_type == "melp":
        return add_collimator_melp(sim, head, debug)
    if collimator_type == "he":
        return add_collimator_he(sim, head, debug)
    col = ["None", "lehr", "melp", "he"]
    fatal(
        f'Cannot build the collimator "{collimator_type}". '
        f"Available collimator types are: {col}"
    )
    return None


def add_collimator_empty(sim, head):
    mm = g4_units.mm
    colli = sim.add_volume("Box", f"{head.name}_no_collimator")
    colli.mother = head.name
    colli.size = [59.7 * mm, 533 * mm, 387 * mm]
    colli.translation = [-96.7324 * mm, 0, 0]
    blue = [0.5, 0.5, 1, 0.8]
    colli.color = blue
    colli.material = head.material
    return colli


def add_collimator_lehr(sim, head, debug):
    mm = g4_units.mm
    name = head.name
    colli = sim.add_volume("Box", f"{name}_collimator")
    colli.mother = name
    colli.size = [24.05 * mm, 533 * mm, 387 * mm]
    colli.translation = [-78.9074 * mm, 0, 0]
    blue = [0.5, 0.5, 1, 0.8]
    colli.color = blue
    colli.material = "Lead"

    """
    #########################################################################
    #
    # 	Type	|	Diameter	|	Septal thickness	|	No. of holes
    # -----------------------------------------------------------------------
    # 	hex		|	1.11 mm		|	0.16 mm 			|	148000
    #
    #	y spacing	= diameter + septal = 1.27 mm
    #	z spacing	= 2 * (diameter + septal) * sin(60) = 2.19970453 mm
    #
    #	y translation	= y spacing / 2 = 0.635 mm
    #	z translation	= z spacing / 2 = 1.09985 mm
    #
    # 	(this translation from 0,0 is split between hole1 and hole2)
    #
    #	Nholes y	= (No. of holes / sin(60))^0.5 = 413.4
    #	Nholes z	= (No. of holes * sin(60))^0.5 = 358
    #
    #########################################################################
    """

    # hexagon
    hole = sim.add_volume("Hexagon", f"{name}_collimator_hole1")
    hole.height = 24.05 * mm
    hole.radius = 0.555 * mm
    hole.material = "G4_AIR"
    hole.mother = colli.name

    # parameterised holes
    size = [1, 414, 175]
    if debug:
        size = [1, 30, 20]
    tr = [0, 1.27 * mm, 2.1997 * mm, 0]
    rot = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    offset = [0, 0.3175 * mm * 2, 0.549925 * mm * 2, 0]
    repeat_colli_hole(sim, hole, size, tr, rot, offset)

    return colli


def add_collimator_melp(sim, head, debug):
    name = head.name
    mm = g4_units.mm

    colli = sim.add_volume("Box", f"{name}_collimator")
    colli.mother = name
    colli.size = [40.64 * mm, 533 * mm, 387 * mm]
    colli.translation = [-87.2024 * mm, 0, 0]
    blue = [0.5, 0.5, 1, 0.8]
    colli.color = blue
    colli.material = "Lead"

    """
    #########################################################################
    #
    # 	Type	|	Diameter	|	Septal thickness	|	No. of holes
    # -----------------------------------------------------------------------
    # 	hex		|	2.94 mm		|	1.14 mm 			|	14000
    #
    #	y spacing	= diameter + septal = 4.08 mm
    #	z spacing	= 2 * (diameter + septal) * sin(60) = 7.06676729 mm
    #
    #	y translation	= y spacing / 2 = 2.04 mm
    #	z translation	= z spacing / 2 = 3.53338 mm
    #
    # 	(this translation from 0,0 is split between hole1 and hole2)
    #
    #	Nholes y	= (No. of holes / sin(60))^0.5 = 127.1448
    #	Nholes z	= (No. of holes * sin(60))^0.5 = 110.111
    #
    #########################################################################
    """

    # hexagon
    hole = sim.add_volume("Hexagon", f"{name}_collimator_hole1")
    hole.height = 40.64 * mm
    hole.radius = 1.47 * mm
    hole.material = "G4_AIR"
    hole.mother = colli.name
    hole.build_physical_volume = False  # FIXME remove

    # parameterised holes
    size = [1, 128, 55]
    if debug:
        size = [1, 30, 20]
    tr = [0, 4.08 * mm, 7.066767 * mm, 0]
    rot = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    offset = [0, 1.02 * mm * 2, 1.76669 * mm * 2, 0]
    repeat_colli_hole(sim, hole, size, tr, rot, offset)

    return colli


def repeat_colli_hole(sim, hole, size, tr, rot, offset):
    holep = RepeatParametrisedVolume(repeated_volume=hole)
    holep.linear_repeat = size
    holep.translation = tr[0:3]
    holep.rotation = rot
    holep.start = [-(x - 1) * y / 2.0 for x, y in zip(size, tr)]
    # do it twice, with the following offset
    holep.offset_nb = 2
    holep.offset = offset
    sim.volume_manager.add_volume(holep)
    return holep


def add_collimator_he(sim, head, debug):
    name = head.name
    mm = g4_units.mm

    """gate.fatal(
        f"the Intevo HE collimator is not implemented yet. Need to move the shielding ..."
    )"""

    colli = sim.add_volume("Box", f"{name}_collimator")
    colli.mother = name
    colli.size = [59.7 * mm, 583 * mm, 440 * mm]
    colli.translation = [-96.7324 * mm, 0, 0]
    blue = [0.5, 0.5, 1, 0.8]
    colli.color = blue
    colli.material = "Lead"

    """
    #########################################################################
    #
    # 	Type	|	Diameter	|	Septal thickness	|	No. of holes
    # -----------------------------------------------------------------------
    # 	hex		|	4.0 mm		|	2.0 mm 				|	8000
    #
    #	y spacing	= diameter + Septal = 6.0 mm
    #	z spacing	= 2 * (diameter + Septal) * sin(60) = 10.39230485 mm
    #
    #	y translation	= y spacing / 2 = 3.0 mm
    #	z translation	= z spacing / 2 = 5.196152423 mm
    #
    # 	(this translation from 0,0 is split between hole1 and hole2)
    #
    #	Nholes y	= (No. of holes / sin(60))^0.5 = 96.11245657
    #	Nholes z	= (No. of holes * sin(60))^0.5 = 83.23582901
    #
    #########################################################################
    """

    # hexagon
    hole = sim.add_volume("Hexagon", f"{name}_collimator_hole1")
    hole.height = 59.7 * mm
    hole.radius = 2.0 * mm
    hole.material = "G4_AIR"
    hole.mother = colli.name

    # parameterised holes
    size = [1, 96, 42]
    if debug:
        size = [1, 30, 20]
    tr = [0, 6 * mm, 10.39230485 * mm, 0]
    rot = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    offset = [0, -1.5 * mm * 2, -2.598076212 * mm * 2, 0]
    repeat_colli_hole(sim, hole, size, tr, rot, offset)

    return colli


def add_crystal(sim, head):
    mm = g4_units.mm
    front_shield_size = 76 * mm * 2
    yellow = [1, 1, 0, 0.9]

    name = head.name
    crystal_sheath = sim.add_volume("Box", f"{name}_crys_sheath")
    crystal_sheath.mother = name
    crystal_sheath.size = [
        0.3048 * mm,  # , 591 * mm, 445 * mm
        head.size[1] - front_shield_size,
        head.size[2] - front_shield_size,
    ]
    crystal_sheath.translation = [-66.73 * mm, 0, 0]
    crystal_sheath.material = "Aluminium"
    crystal_sheath.color = yellow

    crystal = sim.add_volume("Box", f"{name}_crystal")
    crystal.mother = name
    crystal.size = [
        9.5 * mm,  # , 591 * mm, 445 * mm
        head.size[1] - front_shield_size,
        head.size[2] - front_shield_size,
    ]

    crystal.translation = [-61.8276 * mm, 0, 0]
    crystal.material = "NaI"
    crystal.color = yellow

    return crystal


def add_back_compartment(sim, head):
    mm = g4_units.mm
    name = head.name
    back_compartment = sim.add_volume("Box", f"{name}_back_compartment")
    back_compartment.mother = name
    back_compartment.size = [147.5 * mm, 651.0 * mm, 485.0 * mm]
    back_compartment.translation = [16.6724 * mm, 0, 0]
    back_compartment.material = "G4_AIR"  # FIXME strange ?
    green = [0, 1, 0, 1]
    back_compartment.color = green
    return back_compartment


def add_light_guide(sim, back_compartment):
    mm = g4_units.mm
    name = back_compartment.name
    light_guide = sim.add_volume("Box", f"{name}_light_guide")
    light_guide.mother = name
    light_guide.size = [9.5 * mm, 643.0 * mm, 477.1037366 * mm]
    light_guide.translation = [-69.0 * mm, 0, 0]
    light_guide.material = "Glass"
    green = [0, 1, 0, 1]
    light_guide.color = green

    return light_guide


def compute_plane_position_and_distance_to_crystal_OLD(collimator_type):
    """
    (OLD version only used in spect_validation and spect_into_gan)
    """
    temp_sim = Simulation()
    spect, colli, crystal = add_spect_head(
        temp_sim, "spect", collimator_type, debug=True
    )
    pos = get_volume_position_in_head(temp_sim, "spect", f"collimator", "min", axis=0)
    y = get_volume_position_in_head(temp_sim, "spect", "crystal", "center", axis=0)
    crystal_distance = -y
    psd = -spect.size[0] / 2.0 - 1 * g4_units.nm

    print(f"coll min position {pos} mm")
    print(f"crystal distance {crystal_distance} mm")
    print(f"psd {psd} mm")
    return pos, crystal_distance, psd


def get_normal_to_detector():
    return [1, 0, 0]


def add_detection_plane_for_arf(sim, det_name, colli_type, plane_size=None):
    # the plane is in the world coordinate system outside the real spect head box.

    # user plane size only for debug purpose
    mm = g4_units.mm
    if plane_size is None:
        plane_size = [533 * mm, 387 * mm]

    # plane
    nm = g4_units.nm
    detector_plane = sim.add_volume("Box", det_name)
    detector_plane.material = "G4_Galactic"
    detector_plane.color = [1, 0, 0, 1]
    detector_plane.size = [1 * nm, plane_size[0], plane_size[1]]

    # compute the position according to the front shielding
    p = get_geometrical_parameters()
    arf_position = p[colli_type].psd
    detector_plane.translation = [0, -arf_position, 0]

    # DEBUG
    """fake = sim.add_volume("Box", "fake")
    fake.mother = detector_plane
    fake.material = "G4_Galactic"
    fake.color = [1, 1, 0, 1]
    fake.size = [2 * mm, 20 * mm, 10 * mm]
    fake.translation = [0, -plane_size[0] / 2, -plane_size[1] / 2]"""

    # rotate
    rotate_gantry(detector_plane, radius=0, start_angle_deg=0)

    return detector_plane


def add_source_for_arf_training_dataset(
    sim, source_name, activity, detector_plane, min_energy, max_energy
):
    cm = g4_units.cm
    source = sim.add_source("GenericSource", source_name)
    source.particle = "gamma"
    source.activity = activity
    source.position.type = "disc"
    source.position.radius = 5 * cm
    source.position.translation = [0, -20 * cm, 0]
    source.direction.type = "iso"
    source.energy.type = "range"
    source.energy.min_energy = min_energy
    source.energy.max_energy = max_energy
    source.direction.acceptance_angle.volumes = [detector_plane.name]
    source.direction.acceptance_angle.intersection_flag = True

    return source


def add_actor_for_arf_training_dataset(sim, colli_type, ene_win_actor, rr):
    # the detector is in front of the spect head volume, outside
    detector_plane = add_detection_plane_for_arf(sim, "arf_plane", colli_type)

    # arf actor for building the training dataset
    arf = sim.add_actor("ARFTrainingDatasetActor", "ARF (training)")
    arf.energy_windows_actor = ene_win_actor.name
    arf.attached_to = detector_plane.name
    arf.output_filename = f"arf_training_dataset.root"
    arf.russian_roulette = rr
    arf.plane_axis = [1, 2, 0]  # the depth is plane_axis[2] = 0

    return detector_plane, arf


def add_arf_detector(sim, name, colli_type, image_size, image_spacing, pth_filename):
    # create the plane
    det_plane = add_detection_plane_for_arf(sim, name, colli_type)

    # set the position in front of the collimator
    p = get_geometrical_parameters()
    crystal_distance = p[colli_type].crystal_distance

    arf = sim.add_actor("ARFActor", f"{name}_arf")
    arf.attached_to = det_plane.name
    arf.output_filename = f"projection_{arf.name}.mhd"
    arf.batch_size = 1e5
    arf.image_size = image_size
    arf.image_spacing = image_spacing
    arf.verbose_batch = False
    arf.distance_to_crystal = crystal_distance
    arf.pth_filename = pth_filename
    arf.flip_plane = False
    arf.plane_axis = [1, 2, 0]
    arf.gpu_mode = "auto"

    return det_plane, arf


def rotate_gantry(
    head, radius, start_angle_deg, step_angle_deg=1, nb_angle=1, initial_rotation=None
):
    """
    This function rotates the gantry both for 1) conventional spect or 2) ARF plane,
    according to the value of initial_rotation
    """
    # compute the nb translation and rotation
    translations = []
    rotations = []

    if initial_rotation is None:
        initial_rotation = Rotation.from_euler("xz", (180, 90), degrees=True)

    current_angle_deg = start_angle_deg
    for r in range(nb_angle):
        tr = head.translation.copy()
        tr[1] += radius
        t, rot = get_transform_orbiting(tr, "Z", current_angle_deg)
        rot = Rotation.from_matrix(rot)
        rot = rot * initial_rotation
        rot = rot.as_matrix()
        translations.append(t)
        rotations.append(rot)
        current_angle_deg += step_angle_deg

    # set the motion for the SPECT head
    if nb_angle > 1:
        head.add_dynamic_parametrisation(translation=translations, rotation=rotations)

    # we set the initial position in all cases, this allows for check_overlap to be done
    # with the first position
    head.translation = translations[0]
    head.rotation = rotations[0]


def get_pytomography_detector_physics_data(colli_name):
    cm = g4_units.cm
    # create a fake simulation to get the volume information
    sim = Simulation()
    det, colli, crystal = add_spect_head(sim, f"fake", collimator_type=colli_name)
    holep = sim.volume_manager.find_volumes("collimator_hole1_param")[0]
    hole = sim.volume_manager.find_volumes("collimator_hole1")[0]
    d = {
        "hole_shape": 6,
        "hole_diameter": hole.radius * 2 / cm,
        "hole_spacing": holep.translation[1] / cm,
        "collimator_thickness": hole.height / cm,
        "collimator_material": colli.material.lower(),
        "crystal_width": crystal.size[1] / cm,
        "crystal_height": crystal.size[2] / cm,
    }

    return d


def get_default_size_and_spacing():
    mm = g4_units.mm
    return [128, 128], [4.7951998710632 * mm, 4.7951998710632 * mm]


def add_digitizer(
    sim, crystal_name, name=None, size=None, spacing=None, channels=None, filename=None
):
    # default parameters
    mm = g4_units.mm
    if name is None:
        name = crystal_name
    if size is None:
        size = get_default_size_and_spacing()[0]
    if spacing is None:
        spacing = get_default_size_and_spacing()[1]
    if channels is None:
        channels = get_default_energy_windows("tc99m")

    # create the main digitizer chain
    digitizer = Digitizer(sim, crystal_name, name)

    # Singles
    sc = digitizer.add_module("DigitizerAdderActor", f"{name}_singles")
    sc.group_volume = crystal_name
    sc.policy = "EnergyWeightedCentroidPosition"

    # detection efficiency
    # ea = digitizer.add_module("DigitizerEfficiencyActor")
    # ea.efficiency = 0.86481  # FAKE

    # energy blurring
    keV = g4_units.keV
    eb = digitizer.add_module("DigitizerBlurringActor", f"{name}_eblur")
    eb.blur_attribute = "TotalEnergyDeposit"
    eb.blur_method = "InverseSquare"
    eb.blur_resolution = 0.099  # in %
    eb.blur_reference_value = 140.5 * keV
    # alternative :
    # eb.blur_method = "Linear"
    # eb.blur_resolution = 0.13
    # eb.blur_reference_value = 80 * keV
    # eb.blur_slope = -0.09 * 1 / MeV

    # spatial blurring
    # Source: HE4SPECS - FWHM = 3.9 mm or 3.6 mm?
    # FWHM = 2.sigma.sqrt(2ln2) -> sigma = 1.656 mm
    sb = digitizer.add_module("DigitizerSpatialBlurringActor", f"{name}_sblur")
    sb.blur_attribute = "PostPosition"
    # intrinsic spatial resolution at 140 keV for 9.5 mm thick NaI
    sb.blur_fwhm = 3.6 * mm
    sb.keep_in_solid_limits = True
    sb.use_truncated_Gaussian = False

    # energy windows
    cc = digitizer.add_module("DigitizerEnergyWindowsActor", f"{name}_energy_window")
    cc.channels = channels

    # projection
    proj = digitizer.add_module("DigitizerProjectionActor", f"{name}_projection")
    channel_names = [c["name"] for c in channels]
    proj.input_digi_collections = channel_names
    proj.spacing = spacing
    proj.size = size
    proj.write_to_disk = True
    if filename is not None:
        proj.output_filename = filename

    # projection plane: it depends on how the spect device is described
    # here, we need this rotation
    proj.detector_orientation_matrix = Rotation.from_euler(
        "yx", (90, 90), degrees=True
    ).as_matrix()

    return digitizer


def calculate_collimator_acceptance_angle(collimator_type, energy, prob_threshold):
    p = get_geometrical_parameters()
    hole_diameter = p[collimator_type].hole_diameter
    collimator_length = p[collimator_type].collimator_length
    septa_thickness = p[collimator_type].septa_thickness

    mu_lead_cm = get_mu_from_xraylib("Pb", energy)
    print(
        f"collimator_type: {collimator_type}, energy: {energy} keV, mu_lead_cm: {mu_lead_cm} cm-1"
    )
    print(f"collimator septa thicnkess {septa_thickness} mm")
    theta_max = calculate_theta_max_angle(
        hole_diameter, collimator_length, septa_thickness, mu_lead_cm
    )
    print(f"theta_max: {theta_max} deg")

    return theta_max



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
from opengate.managers import Simulation
from opengate.utility import g4_units, g4_best_unit
from opengate.image import get_translation_between_images_center, read_image_info
from opengate.logger import INFO
from opengate.geometry.materials import HounsfieldUnit_to_material


def create_simulation(param):
    """
    param is dict with:
    - output_folder
    - visu
    - number_of_threads
    - ct_image
    - density_tolerance_gcm3
    - table_mat
    - table_density
    - verbose:
    - radionuclide
    - activity_bq
    """
    # create the simulation
    sim = Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = param.visu
    sim.visu_type = "vrml"
    sim.number_of_threads = param.number_of_threads
    sim.verbose_level = INFO
    param.output_folder = pathlib.Path(param.output_folder)
    sim.output_dir = param.output_folder
    sim.progress_bar = True

    # units
    m = g4_units.m
    mm = g4_units.mm
    keV = g4_units.keV
    Bq = g4_units.Bq
    gcm3 = g4_units.g_cm3

    #  change world size
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]

    # CT image
    if sim.visu:
        ct = sim.add_volume("Box", "ct")
        info = read_image_info(param.ct_image)
        ct.size = info.size
        ct.material = "G4_WATER"
        ct.color = [0, 0, 1, 1]
        sim.number_of_threads = 1
    else:
        ct = sim.add_volume("Image", "ct")
        ct.image = param.ct_image
        ct.material = "G4_AIR"  # material used by default
        tol = param.density_tolerance_gcm3 * gcm3
        ct.voxel_materials, materials = HounsfieldUnit_to_material(
            sim, tol, param.table_mat, param.table_density
        )
        if param.verbose:
            print(f'Density tolerance = {g4_best_unit(tol, "Volumic Mass")}')
            print(
                f"Number of materials in the CT : {len(ct.voxel_materials)} materials"
            )
        ct.dump_label_image = param.output_folder / "labels.mhd"

    # some radionuclides choice
    # (user of this function can still change
    # the source in the output sim)
    rad_list = {
        "Lu177": {"Z": 71, "A": 177, "name": "Lutetium 177"},
        "Y90": {"Z": 39, "A": 90, "name": "Yttrium 90"},
        "In111": {"Z": 49, "A": 111, "name": "Indium 111"},
        "I131": {"Z": 53, "A": 131, "name": "Iodine 131"},
    }

    # Activity source from an image
    source = sim.add_source("VoxelSource", "vox")
    source.attached_to = ct.name
    source.particle = "ion"
    source.ion.Z = rad_list[param.radionuclide]["Z"]
    source.ion.A = rad_list[param.radionuclide]["A"]
    source.activity = param.activity_bq * Bq / sim.number_of_threads
    source.image = param.activity_image
    source.direction.type = "iso"
    source.energy.mono = 0 * keV
    # compute the translation to align the source with CT
    # (considering they are in the same physical space)
    source.position.translation = get_translation_between_images_center(
        param.ct_image, param.activity_image
    )

    # cuts
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = True
    sim.physics_manager.set_production_cut("world", "all", 1 * m)
    sim.physics_manager.set_production_cut("ct", "all", 2 * mm)

    # add dose actor (get the same size as the source)
    source_info = read_image_info(param.activity_image)
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "edep.mhd"
    dose.dose_uncertainty.active = True
    dose.dose_squared.active = True
    dose.dose.active = True
    dose.attached_to = ct.name
    dose.size = source_info.size
    dose.spacing = source_info.spacing
    # translate the dose the same way as the source
    dose.translation = source.position.translation
    # set the origin of the dose like the source
    if not sim.visu:
        dose.output_coordinate_system = "attached_to_image"
    dose.hit_type = "random"
    dose.dose_uncertainty.active = True
    dose.dose_squared.active = True
    dose.dose.active = True

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True
    stats.output_filename = "stats.txt"
    stats.stats.write_to_disk = True

    return sim



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from box import Box
import opengate.contrib.spect.ge_discovery_nm670 as gate_spect
import opengate as gate
from opengate.tests import utility
from opengate.sources.utility import get_spectrum

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test046")

    radionuclides = ["Tc99m", "Lu177", "In111", "I131"]
    ref_collimators = ["lehr", "megp", "megp", "hegp"]

    # Test 1
    is_ok = True
    for rad, col in zip(radionuclides, ref_collimators):
        collimator = gate_spect.get_collimator(rad)
        ok = True
        if collimator != col:
            ok = False
        utility.print_test(ok, f"Rad {rad}: the collimator is {col}")
        is_ok = is_ok and ok

    # Test 2
    print()
    sim = gate.Simulation()
    sim.output_dir = paths.output
    digit = Box()
    digit_ns = Box()
    for rad in radionuclides:
        channels = gate.actors.digitizers.get_simplified_digitizer_channels_rad(
            "fake_spect", rad, True
        )
        # cc = gate_spect.add_digitizer_energy_windows(sim, 'fake_crystal', channels)
        digit[rad] = channels
        channels = gate.actors.digitizers.get_simplified_digitizer_channels_rad(
            "fake_spect", rad, False
        )
        # cc = gate_spect.add_digitizer_energy_windows(sim, 'fake_crystal', channels)
        digit_ns[rad] = channels

    print(digit)
    print()
    print(digit_ns)

    # reference
    """outfile = open(paths.output_ref / "t046_digitizer.json", "w")
    json.dump(digit, outfile, indent=4)
    outfile = open(paths.output_ref / "t046_digitizer_ns.json", "w")
    json.dump(digit_ns, outfile, indent=4)"""

    # check
    ref_digit = utility.read_json_file(paths.output_ref / "t046_digitizer.json")
    ref_digit_ns = utility.read_json_file(paths.output_ref / "t046_digitizer_ns.json")
    ok = digit == ref_digit
    utility.print_test(ok, f"Test channels (with scatter): {ok}")
    is_ok = is_ok and ok
    ok = digit_ns == ref_digit_ns
    utility.print_test(ok, f"Test channels (without scatter): {ok}")
    is_ok = is_ok and ok

    # Test 3
    print()
    yields = [0.8907654364665489, 0.18033, 1.8474, 1.0077]
    tolerance = 0.01
    st = f"(tol = {tolerance * 100:.2f} %)"
    i = 0
    for rad in radionuclides:
        rad_spectrum = get_spectrum(rad, "gamma")
        tw = rad_spectrum["weights"].sum()
        tw_d = tw / yields[i] * 100 - 100
        ok = abs(tw_d) <= tolerance * 100
        utility.print_test(
            ok, f"Test yield {rad}: {tw} {yields[i]}: {tw_d:+.2f} % {st} {ok}"
        )
        is_ok = is_ok and ok
        i += 1

    # end
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itk
import numpy as np
from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility
from opengate.contrib.beamlines.ionbeamline import BeamlineModel
from opengate.contrib.tps.ionbeamtherapy import spots_info_from_txt

if __name__ == "__main__":
    # ------ INITIALIZE SIMULATION ENVIRONMENT ----------
    paths = utility.get_default_test_paths(__file__, "gate_test044_pbs", "test059")
    output_path = paths.output
    ref_path = paths.output_ref

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 12365478910
    sim.random_engine = "MersenneTwister"
    sim.output_dir = output_path

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg
    rad = gate.g4_units.rad

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    # nozzle box
    box = sim.add_volume("Box", "box")
    box.size = [500 * mm, 500 * mm, 1000 * mm]
    box.translation = [1148 * mm, 0.0, 0.0]
    box.rotation = Rotation.from_euler("y", -90, degrees=True).as_matrix()
    box.material = "Vacuum"
    box.color = [0, 0, 1, 1]

    # nozzle WET
    nozzle = sim.add_volume("Box", "nozzle")
    nozzle.mother = box.name
    nozzle.size = [500 * mm, 500 * mm, 2 * mm]
    nozzle.material = "G4_WATER"

    # target
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [300 * mm, 310 * mm, 310 * mm]
    phantom.translation = [150 * mm, 0.0, 0.0]
    # phantom.rotation = Rotation.from_euler('x',-90,degrees=True).as_matrix()
    phantom.material = "G4_AIR"
    phantom.color = [0, 0, 1, 1]
    sim.physics_manager.set_max_step_size(phantom.name, 0.8)
    sim.physics_manager.set_user_limits_particles("all")

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"
    sim.physics_manager.set_production_cut("world", "all", 1000 * km)

    # add dose actor
    dose_actor = sim.add_actor("DoseActor", "doseInXYZ")
    dose_actor.output_filename = "testTPSoptics.mhd"
    dose_actor.attached_to = phantom.name
    dose_actor.size = [30, 620, 620]
    dose_actor.spacing = [10.0, 0.5, 0.5]
    dose_actor.hit_type = "random"
    dose_actor.dose.active = True

    # ---------- DEFINE BEAMLINE MODEL -------------
    IR2HBL = BeamlineModel()
    IR2HBL.name = None
    IR2HBL.radiation_types = "ion 6 12"
    # Nozzle entrance to Isocenter distance
    IR2HBL.distance_nozzle_iso = 1300.00  # 1648 * mm#1300 * mm
    # SMX to Isocenter distance
    IR2HBL.distance_stearmag_to_isocenter_x = 6700.00
    # SMY to Isocenter distance
    IR2HBL.distance_stearmag_to_isocenter_y = 7420.00
    # polinomial coefficients
    IR2HBL.energy_mean_coeffs = [11.91893485094217, -9.539517997860457]
    IR2HBL.energy_spread_coeffs = [0.0004790681841295621, 5.253257865904452]
    IR2HBL.sigma_x_coeffs = [-0.00011142901344618727, 2.346946879501544]
    IR2HBL.theta_x_coeffs = [-3.6368814874049214e-07, 0.0003381328996152591]
    IR2HBL.epsilon_x_coeffs = [3.1292233857396716e-06, 0.0004117718840152502]
    IR2HBL.sigma_y_coeffs = [-0.0004009682717802152, 2.0124504979960225]
    IR2HBL.theta_y_coeffs = [-8.437400716390318e-07, 0.000892426821944524]
    IR2HBL.epsilon_y_coeffs = [-8.757558864087579e-08, 0.00250212397239695]

    # --------START PENCIL BEAM SCANNING----------
    # NOTE: HBL means that the beam is coming from -x (90 degree rot around y)
    # nSim = 328935  # particles to simulate per beam
    nSim = 20000
    beam_data_dict = spots_info_from_txt(
        ref_path / "TreatmentPlan4Gate-gate_test59_TP_1_old.txt", "ion 6 12", beam_nr=1
    )
    ntot = beam_data_dict["msw_beam"]
    tps = sim.add_source("TreatmentPlanPBSource", "TP source")
    tps.beam_model = IR2HBL
    tps.n = nSim
    tps.particle = "ion 6 12"
    tps.beam_data_dict = beam_data_dict
    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # start simulation
    sim.run()

    # -------------END SCANNING------------- #
    # print results at the end
    d_fPath = output_path / sim.get_actor("doseInXYZ").dose.get_output_path()
    print(stats)

    # ------ TESTS -------#
    # dose_path = utility.scale_dose(
    #     str(dose.output).replace(".mhd", "_dose.mhd"),
    #     ntot / actual_sim_particles,
    # )

    # SPOT POSITIONS COMPARISON
    # read output and ref
    img_mhd_out = itk.imread(dose_actor.dose.get_output_path())
    img_mhd_ref = itk.imread(
        ref_path / "idc-PHANTOM-air_box-gate_test59_TP_1-PLAN-Physical.mhd"
    )
    data = itk.GetArrayViewFromImage(img_mhd_out)
    data_ref = itk.GetArrayViewFromImage(img_mhd_ref)
    shape = data.shape
    spacing = img_mhd_out.GetSpacing()

    # spot comparison (N.B x and z are inverted in np array!)
    # spots in the plan file
    yz = [
        50,
        -50,
        -50,
        -50,
        -100,
        0,
        100,
        0,
        50,
        50,
        0,
        100,
        -100,
        -100,
        0,
        -100,
        100,
        -100,
        0,
        0,
        -50,
        50,
        -100,
        100,
        100,
        100,
    ]

    yzM = np.array(yz).reshape(int(len(yz) / 2), 2)
    # convert from mm (wrt image center) to voxel
    spot_y = [
        int(y / dose_actor.spacing[1]) + int(dose_actor.size[1] / 2) for y in yzM[:, 0]
    ]
    spot_z = [
        int(z / dose_actor.spacing[1]) + int(dose_actor.size[1] / 2) for z in yzM[:, 1]
    ]

    thresh = 0.105

    # # 1D
    # fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(25, 10))
    # gate.plot_img_axis(ax, img_mhd_out, "z profile", axis="z")
    # #gate.plot_img_axis(ax, img_mhd_out, "x profile", axis="x")
    # gate.plot_img_axis(ax, img_mhd_out, "y profile", axis="y")

    # # fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(25, 10))
    # gate.plot_img_axis(ax, img_mhd_ref, "z ref", axis="z")
    # #gate.plot_img_axis(ax, img_mhd_ref, "x ref", axis="x")
    # gate.plot_img_axis(ax, img_mhd_ref, "y ref", axis="y")

    # fig.savefig(output_path / "dose_profiles_spots.png")

    ok = True
    for i in range(1, shape[2], shape[2] // 3):
        # check i-th slab
        print(f"Airslab nr. {i}")
        # gate.plot2D(data[:, :, i], "2D Edep opengate", show=True)
        # gate.plot2D(data_ref[:, :, i], "2D Edep gate", show=True)
        for y, z in zip(spot_y, spot_z):
            # i = 0
            print(f" ({y:.2f},{z:.2f})")
            # 'cut' the slab around the spot expected in y,z
            w = 30  # cut window's half size
            d_out = data[z - w : z + w, y - w : y + w, i : i + 1]
            d_ref = data_ref[z - w : z + w, y - w : y + w, i : i + 1]
            ok = (
                utility.test_tps_spot_size_positions(
                    d_out, d_ref, spacing, thresh=thresh, abs_tol=0.3
                )
                and ok
            )

    utility.test_ok(ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itk
import os
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility
from opengate.contrib.beamlines.ionbeamline import BeamlineModel
from opengate.contrib.tps.ionbeamtherapy import spots_info_from_txt, TreatmentPlanSource

if __name__ == "__main__":
    # ------ INITIALIZE SIMULATION ENVIRONMENT ----------
    paths = utility.get_default_test_paths(
        __file__, "gate_test044_pbs", output_folder="test059"
    )
    output_path = paths.output
    ref_path = paths.output_ref

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 12365478910
    sim.random_engine = "MersenneTwister"
    sim.output_dir = output_path

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg
    rad = gate.g4_units.rad

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    # nozzle box
    box = sim.add_volume("Box", "box")
    box.size = [500 * mm, 500 * mm, 1000 * mm]
    box.rotation = Rotation.from_euler("x", -90, degrees=True).as_matrix()
    box.translation = [0.0, -1148 * mm, 0.0]
    box.material = "Air"
    box.color = [0, 0, 1, 1]

    # nozzle WET
    nozzle = sim.add_volume("Box", "nozzle")
    nozzle.mother = box.name
    nozzle.size = [500 * mm, 500 * mm, 2 * mm]
    nozzle.material = "G4_WATER"

    # target
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [300 * mm, 310 * mm, 310 * mm]
    phantom.rotation = Rotation.from_euler("z", -90, degrees=True).as_matrix()
    phantom.translation = [0.0, -150 * mm, 0.0]
    phantom.material = "G4_AIR"
    phantom.color = [0, 0, 1, 1]
    sim.physics_manager.set_max_step_size(phantom.name, 0.8)

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"
    sim.physics_manager.set_production_cut("world", "all", 1000 * km)

    # add dose actor
    dose = sim.add_actor("DoseActor", "doseInXYZ")
    dose.output_filename = "TPS_optics_vbl.mhd"
    dose.attached_to = phantom.name
    dose.size = [30, 620, 620]
    dose.spacing = [10.0, 0.5, 0.5]
    dose.hit_type = "random"
    dose.dose.active = True

    # ---------- DEFINE BEAMLINE MODEL -------------
    IR2VBL = BeamlineModel()
    IR2VBL.name = None
    IR2VBL.radiation_types = "ion 6 12"
    # Nozzle entrance to Isocenter distance
    IR2VBL.distance_nozzle_iso = 1300.00  # 1648 * mm#1300 * mm
    # SMX to Isocenter distance
    IR2VBL.distance_stearmag_to_isocenter_x = 6700.00
    # SMY to Isocenter distance
    IR2VBL.distance_stearmag_to_isocenter_y = 7420.00
    # polinomial coefficients
    IR2VBL.energy_mean_coeffs = [11.91893485094217, -9.539517997860457]
    IR2VBL.energy_spread_coeffs = [0.0004790681841295621, 5.253257865904452]
    IR2VBL.sigma_x_coeffs = [-0.00011142901344618727, 2.346946879501544]
    IR2VBL.theta_x_coeffs = [-3.6368814874049214e-07, 0.0003381328996152591]
    IR2VBL.epsilon_x_coeffs = [3.1292233857396716e-06, 0.0004117718840152502]
    IR2VBL.sigma_y_coeffs = [-0.0004009682717802152, 2.0124504979960225]
    IR2VBL.theta_y_coeffs = [-8.437400716390318e-07, 0.000892426821944524]
    IR2VBL.epsilon_y_coeffs = [-8.757558864087579e-08, 0.00250212397239695]

    # --------START PENCIL BEAM SCANNING----------
    # nSim = 328935  # particles to simulate per beam
    nSim = 20000
    tps = sim.add_source("TreatmentPlanPBSource", "TP source")
    tps.beam_model = IR2VBL
    tps.n = nSim
    tps.particle = "ion 6 12"
    tps.plan_path = ref_path / "TreatmentPlan4Gate-gate_test59tps_v.txt"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # create output dir, if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    # start simulation
    sim.run()

    # -------------END SCANNING-------------
    # print results at the end
    print(stats)

    # ------ TESTS -------
    # dose_path = utility.scale_dose(
    #     str(dose.output).replace(".mhd", "_dose.mhd"),
    #     ntot / actual_sim_particles,
    # )

    # SPOT POSITIONS COMPARISON
    # read output and ref
    img_mhd_out = itk.imread(dose.get_output_path("dose"))
    img_mhd_ref = itk.imread(
        ref_path / "idc-PHANTOM-air_box_vbl-gate_test59tps_v-PLAN-Physical.mhd"
    )
    data = itk.GetArrayViewFromImage(img_mhd_out)
    data_ref = itk.GetArrayViewFromImage(img_mhd_ref)
    shape = data.shape
    spacing = img_mhd_out.GetSpacing()

    # spot comparison (N.B x and z are inverted in np array!)
    # spots in the plan file
    yz = [
        0,
        50,
        0,
        0,
        -62.507,
        -21.669,
        -61.951,
        -72.23,
        50,
        -50,
        50,
        0,
        50,
        50,
        -50,
        50,
    ]

    yzM = np.array(yz).reshape(int(len(yz) / 2), 2)
    # convert from mm (wrt image center) to voxel
    spot_y = [int(y / dose.spacing[1]) + int(dose.size[1] / 2) for y in yzM[:, 0]]
    spot_z = [int(z / dose.spacing[1]) + int(dose.size[1] / 2) for z in yzM[:, 1]]

    thresh = 0.105

    # 1D
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(25, 10))
    utility.plot_img_axis(ax, img_mhd_out, "z profile", axis="z")
    utility.plot_img_axis(ax, img_mhd_out, "x profile", axis="x")
    utility.plot_img_axis(ax, img_mhd_out, "y profile", axis="y")

    # fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(25, 10))
    utility.plot_img_axis(ax, img_mhd_ref, "z ref", axis="z")
    utility.plot_img_axis(ax, img_mhd_ref, "x ref", axis="x")
    utility.plot_img_axis(ax, img_mhd_ref, "y ref", axis="y")
    fig.savefig(output_path / "dose_profiles_spots_vbl.png")

    ok = True
    for i in range(1, shape[2], shape[2] // 3):
        # check i-th slab
        print(f"Airslab nr. {i}")
        # gate.plot2D(data[:, :, i], "2D Edep opengate", show=True)
        # gate.plot2D(data_ref[:, :, i], "2D Edep gate", show=True)
        for y, z in zip(spot_y, spot_z):
            # i = 0
            print(f" ({y:.2f},{z:.2f})")
            # 'cut' the slab around the spot expected in y,z
            w = 30  # cut window's half size
            d_out = data[z - w : z + w, y - w : y + w, i : i + 1]
            d_ref = data_ref[z - w : z + w, y - w : y + w, i : i + 1]
            ok = (
                utility.test_tps_spot_size_positions(
                    d_out, d_ref, spacing, thresh=thresh, abs_tol=0.3
                )
                and ok
            )

    utility.test_ok(ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from opengate.actors.coincidences import coincidences_sorter
import uproot
import os
import numpy as np
import sys
import pandas as pd


def main(dependency="test072_coinc_sorter_step1.py"):

    # If output_singles.root does not exist, run a simulation to generate it
    paths = utility.get_default_test_paths(__file__, output_folder="test072")
    root_filename = paths.output / "output_singles.root"
    if not os.path.exists(root_filename):
        print(f"Simulating singles to create {root_filename} ...")
        os.system(f"python {str(paths.current / dependency)}")

    # Prepare arguments for the coincidence sorter
    root_file = uproot.open(root_filename)
    singles_tree = root_file["Singles_crystal"]
    ns = gate.g4_units.nanosecond
    time_window = 3 * ns
    transaxial_plane = "xy"
    policy = "removeMultiples"
    mm = gate.g4_units.mm
    minDistanceXY = 0 * mm
    maxDistanceZ = 32 * mm

    # Run coincidence sorter and return concidences as a pandas DataFrame
    coincidences_pd = coincidences_sorter(
        singles_tree,
        time_window,
        policy,
        minDistanceXY,
        transaxial_plane,
        maxDistanceZ,
        return_type="pd",
    )

    # Run coincidence sorter again, saving coincidences to a root file
    coincidences_sorter(
        singles_tree,
        time_window,
        policy,
        minDistanceXY,
        transaxial_plane,
        maxDistanceZ,
        output_file_path=paths.output / "coincidences.root",
        output_file_format="root",  # Could be ommitted since "root" is the default format
    )
    # Read back the coincidences from the root file and remove the colunm named "index"
    with uproot.open(paths.output / "coincidences.root") as file:
        tree = file["Coincidences"]
        coincidences_from_root = tree.arrays(library="pd").drop(columns=["index"])
    os.remove(paths.output / "coincidences.root")

    # Coincidence sorter output to HDF5 is supported only in Python 3.10 and higher
    if sys.version_info[1] > 9:
        # Run coincidence sorter again, saving coincidences to a HDF5 file
        coincidences_sorter(
            singles_tree,
            time_window,
            policy,
            minDistanceXY,
            transaxial_plane,
            maxDistanceZ,
            output_file_path=paths.output / "coincidences.hdf5",
            output_file_format="hdf5",
        )
        # Read back the coincidences from the HDF5 file
        coincidences_from_hdf5 = pd.read_hdf(paths.output / "coincidences.hdf5")
        os.remove(paths.output / "coincidences.hdf5")

    # Check that the coincidences from the root file and the HDF5 file are equal to the original DataFrame
    try:
        pd.testing.assert_frame_equal(
            coincidences_from_root,
            coincidences_pd,
            check_dtype=False,
            check_categorical=False,
            check_exact=True,
        )
        # Coincidence sorter output to HDF5 is supported only in Python 3.10 and higher
        if sys.version_info[1] > 9:
            pd.testing.assert_frame_equal(
                coincidences_from_hdf5,
                coincidences_pd,
                check_dtype=False,
                check_categorical=False,
                check_exact=True,
            )
        is_ok = True
    except AssertionError:
        is_ok = False

    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from box import Box
import opengate as gate
import opengate_core as g4
from opengate.tests import utility
from opengate.exception import warning


def print_em_parameters(simulation_engine):
    em = g4.G4EmParameters.Instance()
    sec_model = str(em.PIXECrossSectionModel())
    esec_model = str(em.PIXEElectronCrossSectionModel())
    simulation_engine.user_hook_log = Box(
        {
            "bearden": em.FluoDirectory(),
            "pixe_sec_model": sec_model,
            "epixe_sec_model": esec_model,
        }
    )


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__)

    # create simulation
    sim = gate.Simulation()

    # physics
    mm = gate.g4_units.mm
    eV = gate.g4_units.eV
    MeV = gate.g4_units.keV

    # fake source
    source = sim.add_source("GenericSource", "source")
    source.particle = "gamma"
    source.energy.mono = 1 * MeV
    source.direction.type = "iso"
    source.n = 1

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"
    sim.user_hook_after_init = print_em_parameters

    # start simulation
    sim.run(start_new_process=True)
    h = sim.user_hook_log
    print("output", h)
    is_ok = h.bearden == 0 and h.pixe_sec_model == "Empirical"

    # redo with different fluo dir
    print()
    # sim.g4_commands_after_init.append("/process/em/pixeXSmodel ECPSSR_ANSTO")
    sim.g4_commands_after_init.append("/process/em/pixeXSmodel ECPSSR_ANSTO")
    # sim.g4_commands_before_init.append("/process/em/fluoBearden true")
    sim.g4_commands_before_init.append("/process/em/fluoBearden true")
    sim.run(start_new_process=True)
    h = sim.user_hook_log
    print("output", h)
    is_ok = h.bearden == 1 and h.pixe_sec_model == "ECPSSR_ANSTO" and is_ok

    # redo with different fluo dir
    try:
        print()
        sim.g4_commands_after_init = ["/process/em/fluoBearden true"]
        sim.run(start_new_process=True)
        # The above should have caused an exception
        # not OK if it has not.
        is_ok = False
    except:
        warning("This is CORRECT if it throws an exception")

    utility.test_ok(is_ok)



import os
import shutil
import tarfile
import platform
import sys
import requests
from time import sleep
from pathlib import Path

# Data for Geant4
# Geant4 11.3.2
data_packages = {
    "G4NEUTRONHPDATA": "https://cern.ch/geant4-data/datasets/G4NDL.4.7.1.tar.gz",
    "G4LEDATA": "https://cern.ch/geant4-data/datasets/G4EMLOW.8.6.1.tar.gz",
    "G4LEVELGAMMADATA": "https://cern.ch/geant4-data/datasets/G4PhotonEvaporation.6.1.tar.gz",
    "G4RADIOACTIVEDATA": "https://cern.ch/geant4-data/datasets/G4RadioactiveDecay.6.1.2.tar.gz",
    "G4PARTICLEXSDATA": "https://cern.ch/geant4-data/datasets/G4PARTICLEXS.4.1.tar.gz",
    "G4PIIDATA": "https://cern.ch/geant4-data/datasets/G4PII.1.3.tar.gz",
    "G4REALSURFACEDATA": "https://cern.ch/geant4-data/datasets/G4RealSurface.2.2.tar.gz",
    "G4SAIDXSDATA": "https://cern.ch/geant4-data/datasets/G4SAIDDATA.2.0.tar.gz",
    "G4ABLADATA": "https://cern.ch/geant4-data/datasets/G4ABLA.3.3.tar.gz",
    "G4INCLDATA": "https://cern.ch/geant4-data/datasets/G4INCL.1.2.tar.gz",
    "G4ENSDFSTATEDATA": "https://cern.ch/geant4-data/datasets/G4ENSDFSTATE.3.0.tar.gz",
    "G4CHANNELINGDATA": "https://cern.ch/geant4-data/datasets/G4CHANNELING.1.0.tar.gz",
}


def check_g4_data():
    # check if the G4 data folder is there
    dataLocation = get_g4_data_folder()
    if not dataLocation.exists():
        print("Geant4 data folder does not exist.")
        print("I will create it for you here: " + str(dataLocation))
        print("... and download the G4 data.")
        print("This will take a moment.")
        download_g4_data()
        print("")
        print("Done")
        return
    else:
        # Check if the G4 data folder is up to date
        missing_g4_Data = get_missing_g4_data()
        if len(missing_g4_Data) != 0:
            print("\nI will download a fresh G4 dataset for you.")
            print("This will take a moment.")
            download_g4_data(missing_g4_Data)
            if len(get_missing_g4_data()) == 0:
                print("\nGeant4 data has been set up successfully.")
            else:
                print("There is (still) a problem with the Geant4 data.")
                print("Possibly, some data are missing.")


def download_with_resume(url, out, retries=5, delay=10):
    temp_file = str(out) + ".part"
    headers = {}
    if os.path.exists(temp_file):
        file_size = os.path.getsize(temp_file)
        headers["Range"] = f"bytes={file_size}-"

    for attempt in range(retries):
        try:
            with requests.get(url, headers=headers, stream=True) as r:
                r.raise_for_status()
                with open(temp_file, "ab") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
            os.rename(temp_file, str(out))
            print(f"Downloaded {url} successfully.")
            return
        except requests.exceptions.RequestException as e:
            print(f"Download attempt {attempt + 1} failed: {e}")
            if attempt < retries - 1:
                print(f"Retrying in {delay} seconds...")
                sleep(delay)
            else:
                print(f"Exceeded maximum retries for {url}.")
                raise e


# Download Geant4 data:
def download_g4_data(missing_g4_Data=None):
    data_location = get_g4_data_folder()
    data_location.mkdir(parents=True, exist_ok=True)
    folder_names_from_tar = set()

    if missing_g4_Data is None:
        data_packages_needed = list(data_packages.values())
    else:
        data_packages_needed = [
            package
            for g4_data, package in data_packages.items()
            if g4_data in missing_g4_Data
        ]

    for i, package in enumerate(data_packages_needed):
        print(f"\nDownloading {i + 1}/{len(data_packages_needed)} {package}")

        # download the archive (with resume if the connexion failed)
        package_archive = package.split("/")[-1]
        out = os.path.join(data_location, package_archive)
        download_with_resume(package, out)
        # packageArchive = wget.download(package, out=dataLocation)
        package_archive = out

        with tarfile.open(package_archive) as tar:
            # extract the base folder from the tar archive
            # into which the G4 data will be extracted
            extract_to_folder = sorted(tar.getnames(), key=lambda n: len(n))[0]
            folder_names_from_tar.update([extract_to_folder])
            extract_to_path = data_location / extract_to_folder
            # remove the directory into which tar wants extract if it
            # already exists to avoid permission error
            if extract_to_path.exists():
                print(f"\nNeed to extract into {extract_to_path},")
                print("but that directory already exists.")
                print("I need to remove it.")
                shutil.rmtree(extract_to_path)
            print("Extracting the data archive (tar) ...")
            tar.extractall(path=data_location)
            print("done")
        os.remove(package_archive)
    check_for_non_required_files_folders()


def check_for_non_required_files_folders():
    """Check if there are old data folders and inform the user."""
    dataLocation = get_g4_data_folder()
    required_paths = set(get_g4_data_paths().values())
    existing_paths = set([(dataLocation / f) for f in dataLocation.iterdir()])
    outdated_paths = existing_paths.difference(required_paths)
    if len(outdated_paths) > 0:
        print("\n" + 10 * "*")
        print(f"The following files and folders in {dataLocation}")
        print(f"are not required and can be safely deleted:\n")
        for f in outdated_paths:
            print(str(f))
        print("\n" + 10 * "*")


def get_missing_g4_data() -> list:
    dataLocation = get_g4_data_folder()
    required_paths = set(get_g4_data_paths().values())
    existing_paths = set([(dataLocation / f) for f in dataLocation.iterdir()])
    missing_paths = required_paths.difference(existing_paths)
    if len(missing_paths) > 0:
        print("\nSome Geant4 data folder seem to be missing, namely:")
        for p in missing_paths:
            print(str(p))
        return [
            g4_data
            for g4_data, folder in get_g4_data_paths().items()
            if folder in missing_paths
        ]
    else:
        return []


# Return Geant4 data folder:
def get_g4_data_folder() -> Path:
    package_location = Path(__file__).resolve().parent
    return package_location / "geant4_data"


# Return Geant4 data path:
def get_g4_data_paths() -> dict:
    data_location = get_g4_data_folder()
    # 11.3.2
    g4_data_path = {
        "G4NEUTRONHPDATA": data_location / "G4NDL4.7.1",
        "G4LEDATA": data_location / "G4EMLOW8.6.1",
        "G4LEVELGAMMADATA": data_location / "PhotonEvaporation6.1",
        "G4RADIOACTIVEDATA": data_location / "RadioactiveDecay6.1.2",
        "G4PARTICLEXSDATA": data_location / "G4PARTICLEXS4.1",
        "G4PIIDATA": data_location / "G4PII1.3",
        "G4REALSURFACEDATA": data_location / "RealSurface2.2",
        "G4SAIDXSDATA": data_location / "G4SAIDDATA2.0",
        "G4ABLADATA": data_location / "G4ABLA3.3",
        "G4INCLDATA": data_location / "G4INCL1.2",
        "G4ENSDFSTATEDATA": data_location / "G4ENSDFSTATE3.0",
        "G4CHANNELINGDATA": data_location / "G4CHANNELING1.0",
    }
    return g4_data_path


# Set Geant4 data paths:
def set_g4_data_path():
    g4_data_path = get_g4_data_paths()
    for key, value in g4_data_path.items():
        os.environ[key] = str(value)
    s = platform.system()
    g4_lib_folder = None

    if s == "Linux" or s == "Windows":
        g4_lib_folder = Path(__file__).resolve().parent.parent / "opengate_core.libs"
    elif s == "Darwin":
        g4_lib_folder = Path(__file__).resolve().parent / ".dylibs"

    if s == "Windows":
        os.add_dll_directory(str(g4_lib_folder))
    else:
        sys.path.append(str(g4_lib_folder))

    if "LD_LIBRARY_PATH" not in os.environ:
        os.environ["LD_LIBRARY_PATH"] = ""
    os.environ["LD_LIBRARY_PATH"] = (
        str(g4_lib_folder) + ":" + os.environ["LD_LIBRARY_PATH"]
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import numpy as np
from opengate.tests import utility

# Warning: This test is outdated.
# It relies on functionality of the dose actor that is currently not available.


def run_sim(n_thr, c4_ref=None, paths=None):
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 123456789
    sim.number_of_threads = n_thr
    sim.output_dir = paths.output
    Ntotal = 10000 * (30 / n_thr) ** 2
    N_per_trhead = Ntotal / sim.number_of_threads

    # units
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    km = gate.g4_units.km
    MeV = gate.g4_units.MeV

    # add a material database
    # sim.add_material_database(paths.gate_data / "HFMaterials2014.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    # waterbox
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [50 * mm, 100 * mm, 100 * mm]
    phantom.translation = [-25 * mm, 0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMY"
    sim.physics_manager.global_production_cuts.all = 1000 * km

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 80 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    source.position.sigma_x = 2 * mm
    source.position.sigma_y = 2 * mm
    source.position.translation = [0, 0, 0]
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    source.n = N_per_trhead

    dose_size = [100, 1, 1]
    dose_spacing = [0.5 * mm, 100.0 * mm, 100.0 * mm]
    doseActorName_IDD_singleImage = "IDD_singleImage"
    doseActor = sim.add_actor("DoseActor", doseActorName_IDD_singleImage)
    doseActor.output_filename = f"test041-{doseActorName_IDD_singleImage}.mhd"
    doseActor.attached_to = phantom.name
    doseActor.size = dose_size
    doseActor.spacing = dose_spacing
    doseActor.hit_type = "random"
    doseActor.user_output.edep_uncertainty.active = True

    doseActorName_IDD_NthreadImages = "IDD_NthreadImages"
    doseActor = sim.add_actor("DoseActor", doseActorName_IDD_NthreadImages)
    doseActor.output_filename = f"test041-{doseActorName_IDD_NthreadImages}.mhd"
    doseActor.attache_to = phantom.name
    doseActor.size = dose_size
    doseActor.spacing = dose_spacing
    doseActor.hit_type = "random"
    # doseActor.use_more_ram = True
    # doseActor.ste_of_mean = True

    doseActorName_IDD_NthreadImages_unbiased = "IDD_NthreadImages_unbiased"
    doseActor = sim.add_actor("DoseActor", doseActorName_IDD_NthreadImages_unbiased)
    doseActor.output_filename = (
        f"test041-{doseActorName_IDD_NthreadImages_unbiased}.mhd"
    )
    doseActor.attache_to = phantom.name
    doseActor.size = dose_size
    doseActor.spacing = dose_spacing
    doseActor.hit_type = "random"
    # doseActor.use_more_ram = True
    # doseActor.ste_of_mean_unbiased = True

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run(start_new_process=True)

    # print results at the end
    print(stats)

    # ----------------------------------------------------------------------------------------------------------------
    # tests

    doseFpath_IDD_singleImage = str(
        sim.get_actor(doseActorName_IDD_singleImage).get_output_path("edep")
    )
    doseFpath_IDD_NthreadImages = str(
        sim.get_actor(doseActorName_IDD_NthreadImages).get_output_path("edep")
    )
    doseFpath_IDD_NthreadImages_uncert = str(
        sim.get_actor(doseActorName_IDD_NthreadImages).get_output_path(
            "edep_uncertainty"
        )
    )
    doseFpath_IDD_NthreadImages_uncert_unbiased = str(
        sim.get_actor(doseActorName_IDD_NthreadImages_unbiased).get_output_path(
            "edep_uncertainty"
        )
    )
    doseFpath_IDD_singleImage_uncert = str(
        sim.get_actor(doseActorName_IDD_singleImage).get_output_path("edep_uncertainty")
    )

    unused = utility.assert_images(
        doseFpath_IDD_singleImage,
        doseFpath_IDD_NthreadImages,
        stats,
        tolerance=100,
        ignore_value_data2=0,
        axis="x",
    )
    expected_ratio = 1.00
    gate.exception.warning("Test ratio: dose / dose MT cp image for each trhead")
    is_ok = utility.assert_images_ratio(
        expected_ratio,
        doseFpath_IDD_singleImage,
        doseFpath_IDD_NthreadImages,
        abs_tolerance=0.03,
    )
    gate.exception.warning(
        "Test ratio: uncertainty classic / standard error of mean (of each thread)"
    )
    is_ok = is_ok and utility.assert_images_ratio(
        expected_ratio,
        doseFpath_IDD_singleImage_uncert,
        doseFpath_IDD_NthreadImages_uncert_unbiased,
        abs_tolerance=0.07,
        fn_to_apply=lambda x: np.mean(x),
    )
    gate.exception.warning(
        "Test ratio: unbiased standard error / biased standard error = c4 corr factor "
    )
    if c4_ref:
        is_ok = is_ok and utility.assert_images_ratio(
            c4_ref,
            doseFpath_IDD_NthreadImages_uncert_unbiased,
            doseFpath_IDD_NthreadImages_uncert,
            abs_tolerance=0.10,
            fn_to_apply=lambda x: np.mean(x),
        )
    return is_ok


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test041_dose_actor_dose_to_water", "test041"
    )

    is_ok_c4 = []
    is_ok_uncert = []
    n_thrV = [6, 30]
    pass_rates_V = []
    c4_referencesV = {
        2: 0.7978845608,
        3: 0.8862269255,
        4: 0.9213177319,
        5: 0.939985603,
        6: 0.951532862,
        7: 0.959368789,
        8: 0.965030456,
        9: 0.9693107,
        10: 0.972659274,
        16: 0.983284169,
        30: 0.9910802775,
        32: 0.99206349,
        100: 0.997477976,
        1000: 0.999749781,
    }
    for n_thr in n_thrV:
        c4_calc = gate.utility.standard_error_c4_correction(n_thr)
        if n_thr in c4_referencesV:
            c4_ref = c4_referencesV[n_thr]
            print(f"{n_thr = }")
        else:
            print(n_thr, "not in c4_referenceV", c4_referencesV)
            raise ValueError()
        is_ok_c4.append(np.abs(c4_calc / c4_ref - 1) < 0.01)
        N_rep = 3
        is_ok_run = np.zeros(N_rep)
        for j in np.arange(0, N_rep):
            is_ok_current = run_sim(n_thr, c4_ref, paths=paths)
            is_ok_run[j] = is_ok_current
        pass_rate = np.sum(is_ok_run) / N_rep
        pass_rates_V.append(pass_rate)
        if pass_rate >= 0.65:
            is_ok_this_N_thread = 1
        else:
            is_ok_this_N_thread = 0
        print(f"{is_ok_this_N_thread =}")
        is_ok_uncert.append(is_ok_this_N_thread)

    print(f"{pass_rates_V =}")
    is_ok = False
    if all(is_ok_c4):
        is_ok = True
    else:
        print("Failed because of incorrect calculation of c4 correction function")
    print()
    if all(is_ok_uncert):
        is_ok = is_ok and True
    else:
        is_ok = False
        print("Uncertainties not correctly calculated")
        print(is_ok_uncert)
        print(n_thrV)
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import numpy as np
from opengate.tests import utility


def run_sim(N_events: int, N_threads: int, N_voxels: int, paths):
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 123456789
    sim.number_of_threads = N_threads
    sim.output_dir = paths.output

    Ntotal = N_events
    N_per_trhead = int(np.round(Ntotal / sim.number_of_threads))

    # units
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    km = gate.g4_units.km
    MeV = gate.g4_units.MeV

    # add a material database
    # sim.add_material_database(paths.gate_data / "HFMaterials2014.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]
    # world.material = "Vacuum"

    # waterbox
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [50 * mm, 100 * mm, 100 * mm]
    phantom.translation = [-25 * mm, 0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMY"
    sim.physics_manager.global_production_cuts.all = 1000 * km

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 80 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    source.position.sigma_x = 2 * mm
    source.position.sigma_y = 2 * mm
    source.position.translation = [0, 0, 0]
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    source.n = N_per_trhead

    # define actors:
    dose_size = [N_voxels, 1, 1]
    # dose_spacing = [5 * mm, 100.0 * mm, 100.0 * mm]
    dose_spacing = np.divide(phantom.size, dose_size)
    print(dose_spacing)

    # create default/reference image
    doseActorName_IDD_singleImage = "IDD_singleImage"
    doseActor = sim.add_actor("DoseActor", doseActorName_IDD_singleImage)
    doseActor.output_filename = f"test041-{doseActorName_IDD_singleImage}.mhd"
    doseActor.attached_to = phantom.name
    doseActor.size = dose_size
    doseActor.spacing = dose_spacing
    doseActor.hit_type = "random"

    # create test image: img for each thread
    doseActorName_IDD_NthreadImages = "IDD_NthreadImages"
    doseActor = sim.add_actor("DoseActor", doseActorName_IDD_NthreadImages)
    doseActor.output_filename = f"test041-{doseActorName_IDD_NthreadImages}.mhd"
    doseActor.attached_to = phantom.name
    doseActor.size = dose_size
    doseActor.spacing = dose_spacing
    doseActor.hit_type = "random"
    # doseActor.use_more_ram = True

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run(start_new_process=True)

    # print results at the end
    print(stats)

    # ----------------------------------------------------------------------------------------------------------------
    # tests
    doseFpath_IDD_singleImage = sim.get_actor(
        doseActorName_IDD_singleImage
    ).get_output_path("edep")
    doseFpath_IDD_NthreadImages = sim.get_actor(
        doseActorName_IDD_NthreadImages
    ).get_output_path("edep")
    return doseFpath_IDD_singleImage, doseFpath_IDD_NthreadImages, stats


def run_test(doseFpath_IDD_singleImage, doseFpath_IDD_NthreadImages, stat):
    unused = utility.assert_images(
        doseFpath_IDD_singleImage,
        doseFpath_IDD_NthreadImages,
        stat,
        tolerance=100,
        ignore_value_data2=0,
        axis="x",
    )
    expected_ratio = 1.00
    gate.exception.warning("Test ratio: edep in single image vs. Nthread image")
    is_ok = utility.assert_images_ratio(
        expected_ratio,
        doseFpath_IDD_singleImage,
        doseFpath_IDD_NthreadImages,
        abs_tolerance=0.05,
    )
    return is_ok


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test041_dose_actor_dose_to_water", "test041"
    )
    N_threadsV = [8, 2]
    N_voxelsV = [50, 2]
    N_events = 800
    summary_str = []
    is_ok = True
    out_str = ""
    for N_thread in N_threadsV:
        for N_voxel in N_voxelsV:
            fPathRef_singleImage, fPathTest_NThreadImgs, stat = run_sim(
                N_events=N_events, N_threads=N_thread, N_voxels=N_voxel, paths=paths
            )
            is_ok_i = run_test(fPathRef_singleImage, fPathTest_NThreadImgs, stat)
            out_str += f"Test {is_ok_i}, for {N_thread = } and {N_voxel = }" + "\n"

            is_ok = is_ok_i and is_ok

    print(out_str)

    utility.test_ok(is_ok)



import inspect
import colored

import opengate_core as g4
from .logger import logger


class GateDeprecationError(Exception):
    """Raise this if a deprecated feature is used.
    Provide the user with information on how to update their code.
    """


class GateFeatureUnavailableError(Exception):
    """Raise this if a feature is used that is (currently) unavailable.
    Ideally, provide the user with information about alternatives.
    Can be used as temporary workaround during refactorings.
    """


class GateImplementationError(Exception):
    """Raise this if a feature is used that is (currently) unavailable.
    Ideally, provide the user with information about alternatives.
    Can be used as temporary workaround during refactorings.
    """


color_error = colored.fore("red") + colored.style("bold")
color_warning = colored.fore("orange_1")
color_ok = colored.fore("light_blue")


def fatal(s):
    caller = inspect.getframeinfo(inspect.stack()[1][0])
    ss = f"Fatal in {caller.filename} line {caller.lineno}"
    logger.critical(ss)
    logger.critical(s)
    raise Exception(s)


def warning(s):
    logger.warning(s)


def raise_except(s):
    # s = colored.stylize(s, color_error)
    raise Exception(s)


class ExceptionHandler(g4.GateExceptionHandler):
    """
    Geant4 exception handler. Inherit form GateExceptionHandler
    The function 'Notify' will be called by G4 when an exception occurs.
    Should be created after initialization.
    Will be automatically stored in the G4 stateManager
    """

    def Notify(self, originOfException, exceptionCode, severity, description):
        s = f"G4Exception origin: {originOfException}\n"
        s += f"G4Exception code: {exceptionCode}\n"
        s += f"G4Exception severity: {severity}\n"
        s += f"G4Exception: {description}"
        if severity == g4.FatalException or severity == g4.FatalErrorInArgument:
            fatal(s)
        warning(s)
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import uproot
import numpy as np
from opengate.tests import utility

# units
m = gate.g4_units.m
mm = gate.g4_units.mm
cm = gate.g4_units.cm
nm = gate.g4_units.nm
Bq = gate.g4_units.Bq
MeV = gate.g4_units.MeV
deg = gate.g4_units.deg


def is_test_ok(phsp_1, phsp_2):
    f1 = uproot.open(str(phsp_1) + ":PhaseSpace1")
    f2 = uproot.open(str(phsp_2) + ":PhaseSpace1")
    df1 = f1.arrays()
    df2 = f2.arrays()

    # print(df1.data,type(df1))
    for key in df1.fields:
        data_1 = df1[key]
        data_2 = df2[key]
        mean_data_1 = np.mean(data_1)
        mean_data_2 = np.mean(data_2)
        std_dev_data_1 = np.std(data_1, ddof=1)
        std_dev_data_2 = np.std(data_2, ddof=1)

        std_err = np.sqrt(
            (std_dev_data_1 / np.sqrt(len(data_1))) ** 2
            + (std_dev_data_2 / np.sqrt(len(data_2)) ** 2)
        )
        print("data 1 mean for " + key + " = " + str(mean_data_1))
        print("data 2 mean for " + key + " = " + str(mean_data_2))
        print("standard error for " + key + " = " + str(std_err))
        print("")
        if np.abs(mean_data_1 - mean_data_2) / std_err > 4:
            return False
    return True


def add_source(sim, plan_to_attach):
    source = sim.add_source("GenericSource", "GenSource")
    source.particle = "gamma"
    source.n = 10000
    source.position.type = "disc"
    radius = plan_to_attach.rmax
    source.position.radius = radius
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.energy.type = "mono"
    source.energy.mono = 1 * MeV
    source.attached_to = plan_to_attach.name


def add_source_2(sim, plan_to_attach):
    source = sim.add_source("GenericSource", "GenSource")
    source.particle = "gamma"
    source.n = 10000
    source.position.type = "disc"
    radius = plan_to_attach.rmax
    source.position.radius = radius
    source.direction.type = "iso"
    source.energy.type = "mono"
    source.energy.mono = 1 * MeV
    source.attached_to = plan_to_attach.name


def add_phsp_source(sim, plan_to_attach):
    source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    paths = utility.get_default_test_paths(__file__, "", "test060")
    source.phsp_file = paths.output / "test060_phsp_isotropic.root"
    source.isotropic_direction = True
    source.position_key = "PrePositionLocal"
    source.direction_key = "PreDirectionLocal"
    source.global_flag = True
    source.particle = "gamma"
    source.batch_size = 50000
    source.n = 10000
    source.verbose = False
    source.attached_to = plan_to_attach.name
    return source


def add_phsp_sphere_actor(sim, name):
    cm = gate.g4_units.cm
    nm = gate.g4_units.nm
    sphere = sim.volume_manager.add_volume("Sphere", "a_sphere")
    sphere.rmin = 50 * cm
    sphere.rmax = sphere.rmin + 1 * nm

    ta1 = sim.add_actor("PhaseSpaceActor", "PhaseSpace1")
    ta1.attached_to = sphere.name
    ta1.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
    ]
    ta1.output_filename = name


def add_phsp_actor(sim, plan_to_attach):
    ta1 = sim.add_actor("PhaseSpaceActor", "PhaseSpace1")
    ta1.attached_to = plan_to_attach.name
    ta1.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
    ]
    ta1.output_filename = "test060_phsp_isotropic.root"


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test060")

    for i in range(3):
        sim = gate.Simulation()
        # main options
        sim.output_dir = paths.output
        print(sim.output_dir)
        sim.g4_verbose = False
        # sim.visu = True
        sim.visu_type = "vrml"
        sim.check_volumes_overlap = False
        # sim.running_verbose_level = gate.EVENT
        sim.number_of_threads = 1
        sim.random_seed = 987654321

        # units
        m = gate.g4_units.m
        cm = gate.g4_units.cm
        nm = gate.g4_units.nm
        MeV = gate.g4_units.MeV

        ##########################################################################################
        # geometry
        ##########################################################################################
        #  adapt world size
        world = sim.world
        world.size = [1 * m, 1 * m, 1 * m]
        world.material = "G4_Galactic"

        # virtual plane for phase space
        s_plane = sim.add_volume("Tubs", "source_plane")
        # plane.material = "G4_AIR"
        s_plane.material = "G4_Galactic"
        s_plane.rmin = 0
        s_plane.rmax = 30 * cm
        s_plane.dz = 1 * nm  # half height
        # plane.rotation = Rotation.from_euler("xy", [180, 30], degrees=True).as_matrix()
        s_plane.translation = [0 * mm, 0 * mm, 0 * mm]
        s_plane.color = [1, 0, 0, 1]  # red

        plane = sim.add_volume("Tubs", "phsp_actor_plane")
        # plane.material = "G4_AIR"
        plane.material = "G4_Galactic"
        plane.rmin = 0
        plane.rmax = 30 * cm
        plane.dz = 1 * nm  # half height
        # plane.rotation = Rotation.from_euler("xy", [180, 30], degrees=True).as_matrix()
        plane.translation = [0 * mm, 0 * mm, 1 * cm]
        plane.color = [1, 0, 0, 1]  # red

        ##########################################################################################
        # Actors
        ##########################################################################################
        # PhaseSpace Actor

        if i == 0:
            add_source(sim, s_plane)
            add_phsp_actor(sim, plane)
        if i == 1:
            add_source_2(sim, s_plane)
            add_phsp_sphere_actor(sim, "test060_phsp_actor_sphere_ref.root")

        if i == 2:
            add_phsp_source(sim, s_plane)
            add_phsp_sphere_actor(sim, "test060_phsp_actor_sphere_phsp.root")

        sim.run(start_new_process=True)

    is_ok = is_test_ok(
        paths.output / "test060_phsp_actor_sphere_ref.root",
        paths.output / "test060_phsp_actor_sphere_phsp.root",
    )
    utility.test_ok(is_test_ok)
